/*
*
*      AggregPacking : Generator of random aggregate packings in concrete     
*                                                                             
*                           version 0.1 (2014-08-22)                          
* 
*                      Copyright (C)  2014  Jan Stransky                      
* 
*          Czech Technical University, Faculty of Civil Engineering,          
*      Department of Structural Mechanics, 166 29 Prague, Czech Republic      
* 
* AggregPacking is free software: you can redistribute it and/or modify it
* under the terms of the GNU Lesser General Public License as published by the
* Free Software Foundation, either version 3 of the License, or (at your
* option) any later version.
* 
* AggregPacking is distributed in the hope that it will be useful, but WITHOUT
* ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
* FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License
* for more details.
* 
* You should have received a copy of the GNU Lesser General Public License
* along with this program. If not, see <http://www.gnu.org/licenses/>.
*/

#include "aggregate.hpp"
#include "randomizer.hpp"

#ifdef AGGREGPACKING_BOOST_PYTHON
	#include <boost/python.hpp>
#endif

#include <stdexcept>

namespace aggregpacking {


/**********************************************************************
* AGGREGATE
**********************************************************************/
bool
Aggregate::isInsideCube(double s) const {
	for (int axis=0; axis<3; axis++) {
		if (computeMinCoord(axis) < 0. || computeMaxCoord(axis) > s) return false;
	}
	return true;
}

#ifdef AGGREGPACKING_SERIALIZATION
Aggregate*
Aggregate::Load(istream& is) {
	string className;
	is >> className;
	if (className == Sphere::className) {
		Sphere* s = new Sphere();
		is >> *s;
		return s;
	}
	if (className == Ellipsoid::className) {
		Ellipsoid* e = new Ellipsoid();
		is >> *e;
		return e;
	}
	LOG_ERROR("unknown Aggregate type ("+className+")")
}

/// Aggregate to ostream shift operator
ostream& operator<<(ostream& os, const Aggregate& o) {
	const Sphere* s = dynamic_cast<const Sphere*>(&o);
	if (s) {
		os << *s;
		return os;
	}
	const Ellipsoid* e = dynamic_cast<const Ellipsoid*>(&o);
	if (e) {
		os << *e;
		return os;
	}
	LOG_ERROR("unsupported argument type")
}
#endif





/**********************************************************************
* SPHERE
**********************************************************************/
bool
Sphere::containsPoint(const Vector3& p) const {
	return (pos-p).squaredNorm() <= radius*radius;
}

bool
Sphere::containsSphere(const Sphere& s) const {
	if (radius < s.radius) return false;
	return (pos-s.pos).squaredNorm() <= pow(radius-s.radius,2);
}

void
Sphere::fromSize(double s) {
	radius = .5*s;
}

double
Sphere::computeDistanceFrom(const Aggregate& aggreg) const {
	const Sphere* s = dynamic_cast<const Sphere*>(&aggreg);
	if (s) {
		const Sphere& ss = *s;
		return (pos-ss.pos).norm() - (radius+ss.radius);
	}
	return aggreg.computeDistanceFrom(*this);
}

#ifdef AGGREGPACKING_SERIALIZATION
const string Sphere::className = "Sphere";

/// Sphere to ostream shift operator
ostream& operator<<(ostream& os, const Sphere& o) {
	os << Sphere::className << " " << o.number << " " << o.pos << " " << o.radius;
	return os;
}

/// Sphere from istream shift operator
istream& operator>>(istream& is, Sphere& o) {
	is >> o.number >> o.pos >> o.radius;
	return is;
}
#endif




/**********************************************************************
* ELLIPSOID
**********************************************************************/
double Ellipsoid::randomAxesRatio = 0.5;

Ellipsoid::Ellipsoid(const Sphere& s) {
	a = b = c = s.getRadius();
	pos = s.getPos();
	ori = Quaternion(1,0,0,0);
}

double
Ellipsoid::computeRelativeDistanceFrom(const Vector3& p, const Matrix3& bT, const Vector3& pos) {
	// From equation {x | ||bT*(x-pos)|| <= 1.0 }
	return (bT*(p-pos)).norm();
}

double
Ellipsoid::computeRelativeDistanceFrom(const Vector3& p) const {
	return computeRelativeDistanceFrom(p,computeMatrixBT(),pos);
}

bool
Ellipsoid::containsPoint(const Vector3& p, const Matrix3& bT, const Vector3& pos) {
	// From equation {x | ||bT*(x-pos)|| <= 1.0 }
	return (bT*(p-pos)).squaredNorm() <= 1.;
}

bool
Ellipsoid::containsPoint(const Vector3& p) const {
	return Ellipsoid::containsPoint(p,computeMatrixBT(),pos);
}

bool
Ellipsoid::containsSphere(const Sphere& s) const {
#if 0
	Vector3 d = s.getPos() - pos;
	double n = d.norm();
	n += s.getRadius();
	d.normalize();
	d *= n;
	d += pos;
	return containsPoint(d);
#else
	Matrix3 mT = computeMatrixBT();
	Matrix3 m = mT.transpose();
	Matrix3 mTi = computeMatrixBTInversed();
	Matrix3 mi = mTi.transpose();
	Vector3 b = pos;
	//
	Ellipsoid e(s);
	Matrix3 bHat = mi*e.computeMatrixB();
	Matrix3 bHatT = bHat.transpose();
	Vector3 cHat = mT*(e.pos-b);
	Vector3 vS(0,0,0);
	Vector3 vE = Ellipsoid::computeFarthestPointFrom(vS,bHatT,cHat);
	return vE.squaredNorm()<1? true : false;
#endif
}

void
Ellipsoid::fromSize(double s) {
	#define R Randomizer::random()
	b = .5*s;
	a = b*(1.+randomAxesRatio*R);
	c = b*(1.-randomAxesRatio*R);
	Quaternion q = Quaternion(R,R,R,R);
	setOri(q);
	#undef R
}

double
Ellipsoid::computeDistanceFrom(const Aggregate& aggreg) const {
	const Sphere* s = dynamic_cast<const Sphere*>(&aggreg);
	if (s) {
		// treat sphere as a special ellipsoid...
		Ellipsoid e(*s);
		return computeDistanceFrom(e);
	}
	const Ellipsoid* e = dynamic_cast<const Ellipsoid*>(&aggreg);
	if (e) {
		double estimation = computeEstimatedDistanceFrom(*e);
		// quick check
		if (estimation<=0.) return estimation;
		// Having two pointa, originally centers of both ellipsoids ...
		Vector3 v1 = pos, v2=e->pos;
		double dLast, ret = (v1-v2).squaredNorm(), tol=(pos-e->pos).norm()*1e-6;
		int i=0;
		// iterate ...
		do {
			dLast = ret;
			// change points to be the closest points from each other ...
			v1 = computeNearestPointFrom(v2);
			v2 = e->computeNearestPointFrom(v1);
			ret = (v1-v2).squaredNorm();
			// until their distance is almost the same than in the previous step
			// or too many steps is passed and the algorithm should be converged and only due to numerical errors it is not according to tol
		} while (std::abs(dLast-ret) > tol && i++ < 1000);
		return sqrt(ret);
	}
	LOG_ERROR("unsupported argument type")
}

double
Ellipsoid::computeEstimatedDistanceFrom(const Ellipsoid& e) const {
	// transform receiver to unit sphere with center in origin
	Matrix3 mT = computeMatrixBT();
	Matrix3 m = mT.transpose();
	Matrix3 mTi = computeMatrixBTInversed();
	Matrix3 mi = mTi.transpose();
	Vector3 b = pos;
	// use the same transformation on the second Ellipsoid
	Matrix3 bHat = mi*e.computeMatrixB();
	Matrix3 bHatT = bHat.transpose();
	Vector3 cHat = mT*(e.pos-b);
	Vector3 vS(0,0,0);
	// if true, both transformed ellipsoids has (0,0,0) in comon, return some negative value
	if (Ellipsoid::containsPoint(vS,bHatT,cHat)) return -.5*e.getSize();
	// compute vE, nearest point of transformed e from (0,0,0) ...
	Vector3 vE = Ellipsoid::computeNearestPointFrom(vS,bHatT,cHat);
	// and checks its value in terms of unit sphere
	int sign = vE.squaredNorm()>1? 1 : -1;
	// vS is nearest point from vE on the unit sphere
	vS = vE.normalized();
	// transform vE and vS back to original configuration
	vE = b + mTi*vE;
	vS = b + mTi*vS;
	// return true distance of vE and vS with correct sign
	double ret = sign*((vE-vS).norm());
#ifdef AGGREGPACKING_DEBUG
	if (isnan(ret)) {
		 LOG_DEBUG2(mi,mTi);
		 LOG_DEBUG(ret);
		 throw;
	}
#endif
	return ret;
}

Vector3
Ellipsoid::computeNearestPointFrom(const Vector3& p, const Matrix3& bT, const Vector3& pos) {
	// if p is contained by the Ellipsoid, return it
	if (Ellipsoid::containsPoint(p,bT,pos)) return Vector3(p);
	/* minimize (x-p) dot (x-p) with constraint (x-pos)A(x-pos)=1
	solved using lagrange multipliers:
	Lambda(x,lambda) = (x-p)(x-p) + lambda[(x-pos)A(x-pos)-1]
	DLambda/Dlambda = (x-pos)A(x-pos)-1 = 0
	DLambda/Dx = 2(x-p) + lambda*2A(x-pos) = 0

	From the last equation, x is expressed in terms of lambda

	x = (I+lamda*A)^-1(p+lambda*A*pos)

	substrituted to the first equation (with M = (I+lamda*A) and v = p+lambda*A*pos)

	0 = (x-pos)A(x-pos)-1 = (Minv*v-pos)A(Minv*v-pos) - 1

	lambda is solved by Newton's method with 

	f(lambda) = (Minv*v-pos)A(Minv*v-pos) - 1
	Df(lambda)/Dlambda = 2A(Minv*v-pos)(DMinv/Dlambda*v + Minv*Dv/Dlambda)

	where
	DMinv/Dlambda = -Minv(DM/Dlambda)Minv = -Minv A Minv
	Dv/Dlambda = A*pos
	*/
	const Matrix3 I=Matrix3::Identity(), A=bT.transpose()*bT;
	double lambda=0,d,f,df,dd;
	const double tol = (p-pos).norm()*1e-6;
	Matrix3 Mi,dMi;
	Vector3 v,dv,ff;
	int ii;
	ii = 0;
	do {
		Mi = (I+lambda*A).inverse();
		dMi = -Mi*A*Mi;
		v = lambda*A*pos + p;
		dv = A*pos;
		ff = A*(Mi*v-pos);
		f = (Mi*v-pos).dot(ff)-1;
		df = 2*ff.dot(Mi*dv+dMi*v);
		dd = f/df;
		lambda -= dd;
		if (ii++ >= 1000) {
			LOG_ERROR("Newton's method did not converge")
		}
	} while (std::abs(dd) > tol);
	// when lambda is solved, return x=Minv*v
	return Mi*v;
}

Vector3
Ellipsoid::computeNearestPointFrom(const Vector3& p) const {
	return Ellipsoid::computeNearestPointFrom(p,computeMatrixBT(),pos);
}

Vector3
Ellipsoid::computeFarthestPointFrom(const Vector3& p, const Matrix3& bT, const Vector3& pos) {
	const Matrix3 bti = bT.inverse();
	Vector3 x,w,ret;
	double best=0,temp;
	// Generate 1000 ranom points on receiver's surface and choose the best one
	for (int i=0; i<1000; i++) {
		w = Vector3(Randomizer::random(),Randomizer::random(),Randomizer::random());
		w.normalize();
		x = pos + bti*w;
		temp = (x-p).squaredNorm();
		if (temp > best) {
			best = temp;
			ret = x;
		}
	}
	return ret;
}

Vector3
Ellipsoid::computeFarthestPointFrom(const Vector3& p) const {
	return Ellipsoid::computeFarthestPointFrom(p,computeMatrixBT(),pos);
}

double
Ellipsoid::getSize() const {
	// middle value from a,b,c
	if (a>b && a>c) return 2*std::max(b,c);
	if (b>a && b>c) return 2*std::max(a,c);
	return 2*std::max(a,b);
}

void
Ellipsoid::computeProjectionOnCoordAxis(double& s0, double& w, int axis) const {
	/* using definition { x | x = c + B^−T*y, ||y||<=1 } (equal sign in the last expression for extreme case)
	projection of arbitrary point x on line p:x=x0+s*u is s = u*(x-x0)/(u*u)
	if line is coordinate axis (x0=0, u=ei), then s = ei*x = ei*(pos+B^-T*y) = s0+w*y
	s0 = ei*pos, w = ei*B^-T = B-1*ei
	considering ||y||=1, then s_extreme = s0 +- ||w||
	*/
	Matrix3 bi = computeMatrixBInversed();
	// sets v to be ei, unit coordinate vector
	Vector3 v = Vector3::Zero();
	v[axis] = 1.;
	// computes s0 and ||w||
	s0 = v.dot(pos);
	w = (bi*v).norm();
}

double
Ellipsoid::computeMaxCoord(int axis) const {
	double s0,w;
	computeProjectionOnCoordAxis(s0,w,axis);
	return s0 + w;
}

double
Ellipsoid::computeMinCoord(int axis) const {
	double s0,w;
	computeProjectionOnCoordAxis(s0,w,axis);
	return s0 - w;
}

Matrix3
Ellipsoid::computeMatrixSigma() const {
	Matrix3 ret = Matrix3::Zero();
	ret(0,0) = 1./a;
	ret(1,1) = 1./b;
	ret(2,2) = 1./c;
	return ret;
}

Matrix3
Ellipsoid::computeMatrixSigmaInversed() const {
	Matrix3 ret = Matrix3::Zero();
	ret(0,0) = a;
	ret(1,1) = b;
	ret(2,2) = c;
	return ret;
}

Matrix3
Ellipsoid::computeMatrixSigmaSquare() const {
	Matrix3 ret = Matrix3::Zero();
	ret(0,0) = 1./(a*a);
	ret(1,1) = 1./(b*b);
	ret(2,2) = 1./(c*c);
	return ret;
}

Matrix3
Ellipsoid::computeMatrixB() const {
	return ori.toRotationMatrix() * computeMatrixSigma();
}

Matrix3
Ellipsoid::computeMatrixBT() const {
	return computeMatrixSigma() * ori.toRotationMatrix().transpose();
}

Matrix3
Ellipsoid::computeMatrixBInversed() const {
	return computeMatrixSigmaInversed() * ori.toRotationMatrix().transpose();
}

Matrix3
Ellipsoid::computeMatrixBTInversed() const {
	return ori.toRotationMatrix() * computeMatrixSigmaInversed();
}

Matrix3
Ellipsoid::computeMatrixA() const {
	Matrix3 r = ori.toRotationMatrix();
	return r * computeMatrixSigmaSquare() * r.transpose();
}

#ifdef AGGREGPACKING_SERIALIZATION
const string Ellipsoid::className = "Ellipsoid";

/// Ellipsoid to ostream shift operator
ostream& operator<<(ostream& os, const Ellipsoid& o) {
	os << Ellipsoid::className << " " << o.number << " " << o.pos << " " << o.ori << " " << o.a << " " << o.b << " " << o.c;
	return os;
}

/// Ellipsoid from istream shift operator
istream& operator>>(istream& is, Ellipsoid& o) {
	is >> o.number >> o.pos >> o.ori >> o.a >> o.b >> o.c;
	return is;
}
#endif




/**********************************************************************
* PYTHON
**********************************************************************/
#ifdef AGGREGPACKING_BOOST_PYTHON
/// @cond DOXYGEN_EXCLUDE
void py_aggregate() {
	using namespace boost::python;

	class_<Aggregate,boost::noncopyable>("Aggregate",no_init);

	Vector3 (Sphere::*sgetPos1)() = &Sphere::getPos;
	void (Sphere::*ssetPos1)(const Vector3&) = &Sphere::setPos;
	Quaternion (Sphere::*sgetOri)() = &Sphere::getOri;
	class_<Sphere,bases<Aggregate>,boost::noncopyable>("Sphere",init<>())
		.def(init<Vector3>())
		.def(init<double>())
		.def(init<Vector3,double>())
		.add_property("radius",&Sphere::getRadius,&Sphere::setRadius)
		.add_property("pos",sgetPos1,ssetPos1)
		.add_property("ori",sgetOri,&Sphere::setOri)
		.add_property("number",&Sphere::getNumber,&Sphere::setNumber)
		.def("containsPoint",&Sphere::containsPoint)
		.def("containsSphere",&Sphere::containsSphere)
		.def("computeDistanceFrom",&Sphere::computeDistanceFrom)
		.def("computeVolume",&Sphere::computeVolume)
		.def("fromSize",&Sphere::fromSize)
		.def("computeMaxCoord",&Sphere::computeMaxCoord)
		.def("computeMinCoord",&Sphere::computeMinCoord)
		.def("isInsideCube",&Sphere::isInsideCube)
		.add_property("size",&Sphere::getSize)
		.add_property("maxSize",&Sphere::getMaxSize)
		.add_property("volume",&Sphere::computeVolume)
		.def(self == Sphere())
		.def(self != Sphere())
#ifdef AGGREGPACKING_SERIALIZATION
		.def(self_ns::str(self_ns::self))
#endif
		;

	Vector3 (Ellipsoid::*egetPos1)() = &Ellipsoid::getPos;
	void (Ellipsoid::*esetPos1)(const Vector3&) = &Ellipsoid::setPos;
	Quaternion (Ellipsoid::*egetOri)() = &Ellipsoid::getOri;
	bool (Ellipsoid::*econtainsPoint)(const Vector3&) const = &Ellipsoid::containsPoint;
	double (Ellipsoid::*ecomputeRelativeDistanceFrom)(const Vector3&) const = &Ellipsoid::computeRelativeDistanceFrom;
	Vector3 (Ellipsoid::*ecomputeNearestPointFrom)(const Vector3&) const = &Ellipsoid::computeNearestPointFrom;
	Vector3 (Ellipsoid::*ecomputeFarthestPointFrom)(const Vector3&) const = &Ellipsoid::computeFarthestPointFrom;
	class_<Ellipsoid,bases<Aggregate>,boost::noncopyable>("Ellipsoid",init<>())
		.def(init<Vector3>())
		.def(init<double,double,double>())
		.def(init<Vector3,double,double,double>())
		.def(init<const Sphere&>())
		.add_property("a",&Ellipsoid::getA,&Ellipsoid::setA)
		.add_property("b",&Ellipsoid::getB,&Ellipsoid::setB)
		.add_property("c",&Ellipsoid::getC,&Ellipsoid::setC)
		.add_property("randomAxesRatio",&Ellipsoid::getRandomAxesRatio,&Ellipsoid::setRandomAxesRatio)
		.add_property("pos",egetPos1,esetPos1)
		.add_property("ori",egetOri,&Ellipsoid::setOri)
		.add_property("number",&Ellipsoid::getNumber,&Ellipsoid::setNumber)
		.def("computeRelativeDistanceFrom",ecomputeRelativeDistanceFrom)
		.def("containsPoint",econtainsPoint)
		.def("containsSphere",&Ellipsoid::containsSphere)
		.def("computeDistanceFrom",&Ellipsoid::computeDistanceFrom)
		.def("computeVolume",&Ellipsoid::computeVolume)
		.def("fromSize",&Ellipsoid::fromSize)
		.def("computeMaxCoord",&Ellipsoid::computeMaxCoord)
		.def("computeMinCoord",&Ellipsoid::computeMinCoord)
		.def("isInsideCube",&Ellipsoid::isInsideCube)
		.def("computeMatrixB",&Ellipsoid::computeMatrixB)
		.def("computeMatrixBT",&Ellipsoid::computeMatrixBT)
		.def("computeMatrixBInversed",&Ellipsoid::computeMatrixBInversed)
		.def("computeMatrixBTInversed",&Ellipsoid::computeMatrixBTInversed)
		.def("computeMatrixSigma",&Ellipsoid::computeMatrixSigma)
		.def("computeMatrixSigmaInversed",&Ellipsoid::computeMatrixSigmaInversed)
		.def("computeMatrixSigmaSquare",&Ellipsoid::computeMatrixSigmaSquare)
		.def("computeMatrixA",&Ellipsoid::computeMatrixA)
		.def("computeNearestPointFrom",ecomputeNearestPointFrom)
		.def("computeFarthestPointFrom",ecomputeFarthestPointFrom)
		.def("computeEstimatedDistanceFrom",&Ellipsoid::computeEstimatedDistanceFrom)
		.add_property("size",&Ellipsoid::getSize)
		.add_property("maxSize",&Ellipsoid::getMaxSize)
		.add_property("volume",&Ellipsoid::computeVolume)
		.add_property("matrixB",&Ellipsoid::computeMatrixB)
		.add_property("matrixBT",&Ellipsoid::computeMatrixBT)
		.add_property("matrixBInversed",&Ellipsoid::computeMatrixBInversed)
		.add_property("matrixBTInversed",&Ellipsoid::computeMatrixBTInversed)
		.add_property("matrixSigma",&Ellipsoid::computeMatrixSigma)
		.add_property("matrixSigmaInversed",&Ellipsoid::computeMatrixSigmaInversed)
		.add_property("matrixSigmaSquare",&Ellipsoid::computeMatrixSigmaSquare)
		.add_property("matrixA",&Ellipsoid::computeMatrixA)
		.def(self == Ellipsoid())
		.def(self != Ellipsoid())
#ifdef AGGREGPACKING_SERIALIZATION
		.def(self_ns::str(self_ns::self))
#endif
		;
}
/// @endcond
#endif

} // namespace
