/*
*
*      AggregPacking : Generator of random aggregate packings in concrete     
*                                                                             
*                           version 0.1 (2014-08-22)                          
* 
*                      Copyright (C)  2014  Jan Stransky                      
* 
*          Czech Technical University, Faculty of Civil Engineering,          
*      Department of Structural Mechanics, 166 29 Prague, Czech Republic      
* 
* AggregPacking is free software: you can redistribute it and/or modify it
* under the terms of the GNU Lesser General Public License as published by the
* Free Software Foundation, either version 3 of the License, or (at your
* option) any later version.
* 
* AggregPacking is distributed in the hope that it will be useful, but WITHOUT
* ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
* FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License
* for more details.
* 
* You should have received a copy of the GNU Lesser General Public License
* along with this program. If not, see <http://www.gnu.org/licenses/>.
*/

/**
 * @file
 * @brief File containing representation of different aggregate types
 */

#ifndef AGGREGATE_HPP
#define AGGREGATE_HPP

#include "math.hpp"
#ifdef AGGREGPACKING_BOOST_PYTHON
	#include <boost/python.hpp>
	namespace bp = boost::python;
#endif

#ifdef AGGREGPACKING_SERIALIZATION
#include <string>
#include <iostream>
#include <fstream>
using std::string;
using std::istream;
using std::ostream;
using std::ifstream;
using std::ofstream;
#endif


namespace aggregpacking {


class Sphere;

/**********************************************************************
* AGGREGATE
**********************************************************************/
/// Abstract class representing one aggregate particle
/** Containing information about its shape, position and orientation
 */
class Aggregate {
	protected:
		/// Number
		int number;

		/// Position
		Vector3 pos;

		/// Orientation
		Quaternion ori;

	public:
		/// Default constructor
		/** Sets number to -1, pos to (0,0,0) and ori to ((1,0,0),0) */
		Aggregate() : Aggregate(Vector3::Zero(),Quaternion::Identity()) {}

		/// Constructor from position
		/** Number is set to -1 and ori to ((1,0,0),0)
		 * @param pos position
		 */
		Aggregate(const Vector3& pos) : Aggregate(pos,Quaternion::Identity()) {}

		/// Constructor from orientation
		/** Number is set to -1 and pos to (0,0,0)
		 * @param ori orientation
		 */
		Aggregate(const Quaternion& ori) : Aggregate(Vector3::Zero(),ori) {}

		/// Constructor from position and orientation
		/** Number is set to -1
		 * @param pos position
		 * @param ori orientation
		 */
		Aggregate(const Vector3& pos, const Quaternion& ori) : number(-1), pos(pos), ori(ori) {}

		/// Equality operator
		/** Intended for Python
		 * @param other instance to compare with
		 */
		bool operator==(const Aggregate& other) const { return &other == this; }

		/// Inequality operator
		/** Intended for Python
		 * @param other instance to compare with
		 */
		bool operator!=(const Aggregate& other) const { return &other != this; }

		/// Ckecks, whether given point is contained by receiver
		/** @param p point coordinates
		 * @return true if receiver contains given point, false otherwise
		 */
		virtual bool containsPoint(const Vector3& p) const = 0;

		/// Ckecks, whether given sphere is contained by receiver
		/** @param s given sphere object
		 * @return true if receiver contains entirely given sphere, false otherwise
		 */
		virtual bool containsSphere(const Sphere& s) const = 0;

		/// Computes distance of receiver from given Aggregate
		/** Positive when the aggregates are separate, negative whan they overlap
		 * @param aggreg Aggregate object to compute distance from
		 * @return distance between receiver and given Aggregate
		 */
		virtual double computeDistanceFrom(const Aggregate& aggreg) const = 0;

		/// Computes maximum coordinate along given axis
		/** @param axis 0=x, 1=y, 2=z
		 */
		virtual double computeMaxCoord(int axis) const = 0;

		/// Computes minimum coordinate along given axis
		/** @param axis 0=x, 1=y, 2=z
		 */
		virtual double computeMinCoord(int axis) const = 0;

		/// Determines if receiver is entirely inside cube. The cube is in positive octant and has one corner in origin
		/** @param s cube size
		 * @return true if receiver is entirely contained by a cube with given size, false otherwise
		 */
		bool isInsideCube(double s) const;

		/// Computes volume of receiver
		virtual double computeVolume() const = 0;

		/// Adjust receiver dimensions accodring to given size
		/** @param s given size
		*/
		virtual void fromSize(double s) = 0;

		/// Returns size of receiver (what sieve opening it would normally pass)
		virtual double getSize() const = 0;

		/// Returns maximum size of receiver
		virtual double getMaxSize() const = 0;

		//static bool compare(Aggregate* a1, Aggregate* a2) { return a2->getSize() < a1->getSize(); }
		/// Getter for number
		int getNumber() { return number; }

		/// Setter for number
		/** @param n new number
		*/
		void setNumber(int n) { number = n; }

		/// Getter for position (by reference)
		const Vector3& getPos() const { return pos; }

		/// Getter for position  (by value)
		Vector3 getPos() { return pos; }

		/// Setter for position
		/** @param p new position
		*/
		void setPos(const Vector3& p) { pos = p; }

		/// Setter for position (from individual components)
		/** @param x x coordinate
		 *  @param y y coordinate
		 *  @param z z coordinate
		 */
		void setPos(double x, double y, double z) {
			pos.x() = x;
			pos.y() = y;
			pos.z() = z;
		}

		/// Getter for orientation (by reference)
		const Quaternion& getOri() const { return ori; }

		/// Getter for orientation (by value)
		Quaternion getOri() { return ori; }

		/// Setter for orientation
		/** @param q new orientation
		*/
		void setOri(const Quaternion& q) { ori = q; ori.normalize(); }

#ifdef AGGREGPACKING_SERIALIZATION
		/// Returns string representation of class name
		virtual const string& getClassName() = 0;

		/// Create and return a new aggregate from given istream
		/** @param is input stream
		*/
		static Aggregate* Load(istream& is);

		/// Aggregate to ostream shift operator
		friend ostream& operator<<(ostream& os, const Aggregate& o);

#endif

		/// Auxiliary function to get x coordinate from Aggregate reference (used by Octree)
		friend double get_x(const Aggregate& a) { return a.pos.x(); }

		/// Auxiliary function to get y coordinate from Aggregate reference (used by Octree)
		friend double get_y(const Aggregate& a) { return a.pos.y(); }

		/// Auxiliary function to get z coordinate from Aggregate reference (used by Octree)
		friend double get_z(const Aggregate& a) { return a.pos.z(); }

		/// Auxiliary function to get (max) size from Aggregate reference (used by Octree)
		friend double get_size(const Aggregate& a) { return a.getMaxSize(); }

		/// Auxiliary function to get x coordinate from Aggregate pointer (used by Octree)
		friend double get_x(const Aggregate* a) { return a->pos.x(); }

		/// Auxiliary function to get y coordinate from Aggregate pointer (used by Octree)
		friend double get_y(const Aggregate* a) { return a->pos.y(); }

		/// Auxiliary function to get z coordinate from Aggregate pointer (used by Octree)
		friend double get_z(const Aggregate* a) { return a->pos.z(); }

		/// Auxiliary function to get (max) size from Aggregate pointer (used by Octree)
		friend double get_size(const Aggregate* a) { return a->getMaxSize(); }
};




/**********************************************************************
* SPHERE
**********************************************************************/
/// Representation of spherical aggregate
/**
 * Spherical particles are defined as a set of 3D space points \f$\mathbf{x}\f$:
 * \f[\{ \mathbf{x}\ |\ ||\mathbf{x}-\mathbf{position}|| \le radius \}\f]
 */
class Sphere : public Aggregate {
	private:
		/// radius
		double radius;

	public:
		/// Default constructor
		Sphere() : Sphere(Vector3::Zero(),0.) {}

		/// Constructor from position
		/** @param p position
		 */
		Sphere(const Vector3& p) : Sphere(p,0.) {}

		/// Constructor from radius
		/** @param r radius
		 */
		Sphere(double r) : Sphere(Vector3::Zero(),r) {}

		/// Constructor from position and radius
		/** @param p position
		 *  @param r radius
		 */
		Sphere(const Vector3& p, double r) : Aggregate(p), radius(r) {}

		/// Getter for radius
		double getRadius() const { return radius; }

		/// Setter for radius
		/** @param r new radius
		 */
		void setRadius(double r) { radius = r; }

		virtual bool containsPoint(const Vector3& p) const;

		virtual bool containsSphere(const Sphere& s) const;

		virtual double computeDistanceFrom(const Aggregate& aggreg) const;

		virtual double computeVolume() const { return 4./3.*PI*pow(radius,3); }

		virtual void fromSize(double s);

		virtual double computeMaxCoord(int axis) const { return pos[axis] + radius; };

		virtual double computeMinCoord(int axis) const { return pos[axis] - radius; };

		virtual double getSize() const { return 2*radius; }

		virtual double getMaxSize() const { return 2*radius; }

		/// String representation of Sphere class name
		const static string className;

#ifdef AGGREGPACKING_SERIALIZATION
		/// Returns string representation of Sphere class name
		virtual const string& getClassName() { return Sphere::className; }

		/// Sphere to ostream shift operator
		friend ostream& operator<<(ostream& os, const Sphere& o);

		/// Sphere from istream shift operator
		friend istream& operator>>(istream& is, Sphere& o);
#endif
};



/**********************************************************************
* ELLIPSOID
**********************************************************************/
/// Representation of ellipsoidal aggregate
/**
 * Ellipsoidal particles are defined as a set of 3D space points \f$\mathbf{x}\f$:
 * \f[\{ \mathbf{x}\ |\ (\mathbf{x}-\mathbf{c})^\mathsf{T}\mathbf{BB}^\mathsf{T}(\mathbf{x}-\mathbf{c}) \le 1 \}\f]
 *
 * \f$\mathbf{c}\f$ is the ellipsoid center. \f$\mathbf{B} = \mathbf{\Sigma}\mathbf{R}^\mathsf{T}\f$ is a matrix defining the ellipsoid. From other possible definitions (see below) it is used most often. \f$\mathbf{\Sigma}\f$ is a diagonal matrix with \f$[\Sigma]_{(ii)}=1/a_i\f$ (\f$a_i\f$ is a length of \f$i\f$th semiaxis) and \f$\mathbf{R}\f$ is an orthogonal rotational matrix (matrix representation of rotation quaternion) with columns equal to rotated coordinate axes.
 *
 * Alternative definition with \f$\mathbf{A}=\mathbf{BB}^\mathsf{T}=\mathbf{R\Sigma}^2\mathbf{R}^\mathsf{T}\f$ is also possible or substituting \f$\mathbf{y}=\mathbf{B}^\mathsf{T}(\mathbf{x}-\mathbf{c})\f$ one can define:
 * \f[\{ \mathbf{x}\ |\ ||\mathbf{B}^\mathsf{T}(\mathbf{x}-\mathbf{c})|| \le 1 \}\f]
 * \f[\{ \mathbf{x}\ |\ (\mathbf{x}-\mathbf{c})^\mathsf{T}\mathbf{A}^\mathsf{T}(\mathbf{x}-\mathbf{c}) \le 1 \}\f]
 * \f[\{ \mathbf{x}\ |\ (\mathbf{x}-\mathbf{c})^\mathsf{T}\mathbf{R\Sigma}^2\mathbf{R}^\mathsf{T}(\mathbf{x}-\mathbf{c}) \le 1 \}\f]
 * \f[\{ \mathbf{x}\ |\ \mathbf{x}=\mathbf{c}+\mathbf{B}^{-\mathsf{T}}\mathbf{y}, ||\mathbf{y})|| \le 1 \}\f]
 */
class Ellipsoid : public Aggregate {
	private:
		/// 1st semiaxis lengths
		double a;

		/// 2nd semiaxis lengths
		double b;

		/// 3rd semiaxis lengths
		double c;

		/// Random axes ratio. Given semiaxis b, a=b*(1+randomAxesRatio*R), c=b*(1-randomAxesRatio*R), R is a random number in range (0,1)
		/** default value is 0.5
		 */
		static double randomAxesRatio;

	public:
		/// Default constructor
		Ellipsoid() : Ellipsoid(Vector3::Zero(),0.,0.,0.) {}

		/// Constructor from position
		/** @param p position
		 */
		Ellipsoid(const Vector3& p) : Ellipsoid(p,0.,0.,0.) {}

		/// Constructor from semiaxes
		/** @param a 1st semiaxis length
		 *  @param b 2nd semiaxis length
		 *  @param c 3rd semiaxis length
		 */
		Ellipsoid(double a, double b, double c) : Ellipsoid(Vector3::Zero(),a,b,c) {}

		/// Constructor from position and semiaxes
		/** @param p position
		 *  @param a 1st semiaxis length
		 *  @param b 2nd semiaxis length
		 *  @param c 3rd semiaxis length
		 */
		Ellipsoid(const Vector3& p,  double a,  double b,  double c) : Aggregate(p), a(a), b(b), c(c) {}

		/// Constructor from Sphere
		/** @param s given Sphere object
		 */
		Ellipsoid(const Sphere& s);

		/// Getter for 1st semiaxis
		double getA() { return a; }

		/// Getter for 2nd semiaxis
		double getB() { return b; }

		/// Getter for 3rd semiaxis
		double getC() { return c; }

		/// Setter for 1st semiaxis
		/** a new semiaxis length
		 */
		void setA(double a) { this->a = a; }

		/// Setter for 2nd semiaxis
		/** b new semiaxis length
		 */
		void setB(double b) { this->b = b; }

		/// Setter for 3rd semiaxis
		/** c new semiaxis length
		 */
		void setC(double c) { this->c = c; }

		/// Getter for randomAxesRatio
		double getRandomAxesRatio() { return randomAxesRatio; }
		
		/// Setter for randomAxesRatio
		void setRandomAxesRatio(double v) { randomAxesRatio = v; }

		/// Computes relative distance between gvien point and given Ellipsoid. Relative distance means in direction pos--p, the value is 1 on surface of the Ellipsoid, in the center is 0, inside Ellipsoid (0,1) and outside Ellipsoid >1
		/** @param p given point
		 *  @param bT \f$\mathbf{R}^\mathsf{T}\f$ matrix of given Ellipsoid
		 *  @param pos position of given Ellipsoid
		 */
		static double computeRelativeDistanceFrom(const Vector3&p, const Matrix3& bT, const Vector3& pos);

		/// Computes relative distance between gvien point and given Ellipsoid. Relative distance means in direction pos--p, the value is 1 on surface of the Ellipsoid, in the center is 0, inside Ellipsoid (0,1) and outside Ellipsoid >1
		/** @param p given point
		 */
		double computeRelativeDistanceFrom(const Vector3& p) const;

		/// Auxiliary structure to test if point lies inside given Ellipsoid
		/** @param p given point
		 *  @param bT \f$\mathbf{R}^\mathsf{T}\f$ matrix of given Ellipsoid
		 *  @param pos position of given Ellipsoid
		 */
		static bool containsPoint(const Vector3&p, const Matrix3& bT, const Vector3& pos);

		virtual bool containsPoint(const Vector3& p) const;

		virtual bool containsSphere(const Sphere& s) const;

		virtual double computeDistanceFrom(const Aggregate& aggreg) const;

		virtual double computeVolume() const { return 4./3.*PI*a*b*c; }

		/// Adjust receiver dimensions accodring to given size. Sets b=0.5*s, a=b*(1+0.5*random), c=b*(1-0.5*random)
		/** @param s given size
		*/
		virtual void fromSize(double s);

		/// Computes parameters of its projction onto coordinate axis (such that the projection is s0+-w)
		/** @param s0 projection of receiver's center
		 *  @param w extreme projection length
		 *  @param axis x=0, y=1, z=2
		 */
		void computeProjectionOnCoordAxis(double& s0, double& w, int axis) const;

		virtual double computeMaxCoord(int axis) const;

		virtual double computeMinCoord(int axis) const;

		/// Returns size of receiver (what sieve opening it would normally pass), i.e. the middle value from a,b,c
		virtual double getSize() const;

		virtual double getMaxSize() const { return 2*std::max(a,std::max(b,c)); }

		/// Computes matrix \f$\mathbf{B} = \mathbf{R\Sigma}\f$
		Matrix3 computeMatrixB() const;

		/// Computes matrix \f$\mathbf{B}^\mathsf{T} = \mathbf{\Sigma R}^\mathsf{T}\f$
		Matrix3 computeMatrixBT() const;

		/// Computes matrix \f$\mathbf{B}^{-1} = \mathbf{\Sigma}^{-1}\mathbf{R}^\mathsf{T}\f$
		Matrix3 computeMatrixBInversed() const;

		/// Computes matrix \f$\mathbf{B}^{-\mathsf{T}} = \mathbf{R\Sigma}^{-1}\f$
		Matrix3 computeMatrixBTInversed() const;

		/// Computes matrix \f$\mathbf{\Sigma}\f$, a diagonal matrix with \f$[\Sigma]_{(ii)}=1/a_i\f$, where \f$a_i\f$ is \f$i\f$th smiaxis length
		Matrix3 computeMatrixSigma() const;

		/// Computes matrix \f$\mathbf{\Sigma}^{-1}\f$, a diagonal matrix with \f$[\Sigma]_{(ii)}=a_i\f$, where \f$a_i\f$ is \f$i\f$th smiaxis length
		Matrix3 computeMatrixSigmaInversed() const;

		/// Computes matrix \f$\mathbf{\Sigma\Sigma}\f$
		Matrix3 computeMatrixSigmaSquare() const;

		/// Computes matrix \f$\mathbf{A}=\mathbf{BB}^\mathsf{T}=\mathbf{R\Sigma}^2\mathbf{R}^\mathsf{T}\f$
		Matrix3 computeMatrixA() const;

		/// Computes nearest point of Ellipsoid to given point. Returns given point if it lies inside the Ellipsoid, returns point on its surface otherwise
		/** @param p given point
		 *  @param bT \f$\mathbf{R}^\mathsf{T}\f$ matrix of given Ellipsoid
		 *  @param pos position of given Ellipsoid
		 */
		static Vector3 computeNearestPointFrom(const Vector3& p, const Matrix3& bT, const Vector3& pos);

		/// Computes nearest point of receiver to given point. Returns given point if it lies inside the receiver, returns point on its surface otherwise
		/** @param p given point
		 */
		Vector3 computeNearestPointFrom(const Vector3& p) const;

		/// Computes farthest point of Ellipsoid to given point. Returns point on its surface. The computation is based on Monte-Carlo random sampling
		/** @param p given point
		 *  @param bT \f$\mathbf{R}^\mathsf{T}\f$ matrix of given Ellipsoid
		 *  @param pos position of given Ellipsoid
		 */
		static Vector3 computeFarthestPointFrom(const Vector3& p, const Matrix3& bT, const Vector3& pos);

		/// Computes farthest point of receiver to given point. Returns point on its surface. The computation is based on Monte-Carlo random sampling
		/** @param p given point
		 */
		Vector3 computeFarthestPointFrom(const Vector3& p) const;

		/// Computes estimated distance between receiver and given Ellipsoid. Mainly used as a quick check if two Ellipsoids overlap (return < 0) or not (return > 0)
		/** @param e given Ellipsoid
		 */
		double computeEstimatedDistanceFrom(const Ellipsoid& e) const;

		/// String representation of Ellipsoid class name
		const static string className;

#ifdef AGGREGPACKING_SERIALIZATION
		/// Returns string representation of Ellipsoid class name
		virtual const string& getClassName() { return Ellipsoid::className; }

		/// Ellipsoid to ostream shift operator
		friend ostream& operator<<(ostream& os, const Ellipsoid& o);

		/// Ellipsoid from istream shift operator
		friend istream& operator>>(istream& is, Ellipsoid& o);
#endif
};



/**********************************************************************
* PYTHON
**********************************************************************/
#ifdef AGGREGPACKING_BOOST_PYTHON
/// @cond DOXYGEN_EXCLUDE
void py_aggregate();
/// @endcond
#endif

} // namespace
#endif
