import subprocess
import os
from math import pow,sqrt

def text2vtk(inFileName,outFileName):
	"""Converts text file (created by :yref:`yade.export.textExt` function) into vtk file.
	See :ysrc:`examples/test/paraview-spheres-solid-section/export_text.py` example

	:param str inFileName: name of input text file
	:param str outFileName: name of output vtk file
	"""
	fin  = open(inFileName)
	fout = open(outFileName,'w')
	lastLine = None
	line = '#'
	while line.startswith('#'):
		lastLine = line
		line = fin.readline()
	columns = lastLine.split()[5:]
	data = [line.split() for line in fin]
	fin.close()
	n = len(data)
	fout.write('# vtk DataFile Version 3.0.\ncomment\nASCII\n\nDATASET POLYDATA\nPOINTS %d double\n'%(n))
	for d in data:
		if len(d)<4: print d
	fout.writelines('%s %s %s\n'%(d[0],d[1],d[2]) for d in data)
	fout.write("\nPOINT_DATA %d\nSCALARS radius double 1\nLOOKUP_TABLE default\n"%(n))
	fout.writelines('%s\n'%(d[3]) for d in data)
	for i,c in enumerate(columns):
		fout.write("\nSCALARS %s double 1\nLOOKUP_TABLE default\n"%(c))
		fout.writelines('%s\n'%(d[4+i]) for d in data)
	fout.close()

def text2vtkSection(inFileName,outFileName,point,normal=(1,0,0)):
	"""Converts section through spheres from text file (created by :yref:`yade.export.textExt` function) into vtk file.
	See :ysrc:`examples/test/paraview-spheres-solid-section/export_text.py` example

	:param str inFileName: name of input text file
	:param str outFileName: name of output vtk file
	:param Vector3|(float,float,float) point: coordinates of a point lying on the section plane
	:param Vector3|(float,float,float) normal: normal vector of the section plane
	"""
	from math import sqrt
	norm = sqrt(pow(normal[0],2)+pow(normal[1],2)+pow(normal[2],2))
	normal = (normal[0]/norm,normal[1]/norm,normal[2]/norm)
	#
	def computeD(point,normal):
		# from point and normal computes parameter d in plane equation ax+by+cz+d=0
		return -normal[0]*point[0] - normal[1]*point[1] - normal[2]*point[2]
	def computeDistanceFromPlane(dat,point,normal,d=None):
		# computes distance of sphere dat from plane (point,normal)
		x,y,z = computeProjectionOnPlane(dat,point,normal,d)
		cx,cy,cz = dat[0],dat[1],dat[2]
		return sqrt(pow(x-cx,2)+pow(y-cy,2)+pow(z-cz,2))
	def computeProjectionOnPlane(self,point,normal,d=None):
		# computes projection of sphere dat on plane (point,normal)
		if d is None:
			d = computeD(point,normal)
		nx,ny,nz = normal[0],normal[1],normal[2]
		cx,cy,cz = dat[0],dat[1],dat[2]
		t = (-d-nx*cx-ny*cy-nz*cz) / (nx*nx+ny*ny+nz*nz)
		x,y,z = cx+t*nx, cy+t*ny, cz+t*nz
		return x,y,z
	#
	fin  = open(inFileName)
	lastLine = None
	line = '#'
	while line.startswith('#'):
		lastLine = line
		line = fin.readline()
	columns = lastLine.split()[4:]
	data = [[float(w) for w in line.split()] for line in fin]
	fin.close()
	#
	d = computeD(point,normal)
	circs = []
	for dat in data:
		r = dat[3]
		dst = computeDistanceFromPlane(dat,point,normal,d)
		if dst > r:
			continue
		x,y,z = computeProjectionOnPlane(dat,point,normal,d)
		rNew = sqrt(r*r-dst*dst)
		dNew = [x,y,z,rNew,r]
		dNew.extend(dat[4:])
		circs.append(dNew)
	n = len(circs)
	fout = open(outFileName,'w')
	fout.write('# vtk DataFile Version 3.0.\ncomment\nASCII\n\nDATASET POLYDATA\nPOINTS %d double\n'%(n))
	fout.writelines('%g %g %g\n'%(c[0],c[1],c[2]) for c in circs)
	fout.write("\nPOINT_DATA %d\nSCALARS radius double 1\nLOOKUP_TABLE default\n"%(n))
	fout.writelines('%g\n'%(c[3]) for c in circs)
	fout.write("\nSCALARS radiusOrig double 1\nLOOKUP_TABLE default\n")
	fout.writelines('%g\n'%(c[4]) for c in circs)
	fout.write("\nVECTORS normal double\n")
	fout.writelines("%g %g %g\n"%normal for i in circs)
	for i,c in enumerate(columns):
		fout.write("\nSCALARS %s double 1\nLOOKUP_TABLE default\n"%(c))
		fout.writelines('%s\n'%(c[4+i]) for c in circs)
	fout.close()

def vtk():
	text2vtk(DAT,SPH)
	text2vtkSection(DAT,SEC,(25e-3,0,0),(1,0,0))
	subprocess.Popen([
		'pvbatch','sph-vtk.py',
		'--sph',SPH,
		'--sec',SEC,
		'--out',PNG,
	]).wait()

DAT = '/tmp/sph.dat'
SPH = '/tmp/sph.vtk'
SEC = '/tmp/sphs.vtk'
PNG = 'sph.png'

if __name__ == "__main__":
	vtk()
	from postproutils import gnuplot
	gnuplot('sph-graph.png','sph.dat','sph-theor.dat')
