/*
*
*      AggregPacking : Generator of random aggregate packings in concrete     
*                                                                             
*                           version 0.1 (2014-08-22)                          
* 
*                      Copyright (C)  2014  Jan Stransky                      
* 
*          Czech Technical University, Faculty of Civil Engineering,          
*      Department of Structural Mechanics, 166 29 Prague, Czech Republic      
* 
* AggregPacking is free software: you can redistribute it and/or modify it
* under the terms of the GNU Lesser General Public License as published by the
* Free Software Foundation, either version 3 of the License, or (at your
* option) any later version.
* 
* AggregPacking is distributed in the hope that it will be useful, but WITHOUT
* ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
* FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License
* for more details.
* 
* You should have received a copy of the GNU Lesser General Public License
* along with this program. If not, see <http://www.gnu.org/licenses/>.
*/

#include "heterogeometry.hpp"

#include <algorithm>
#include <stdexcept>

#ifdef AGGREGPACKING_BOOST_PYTHON
	#include <boost/python.hpp>
#endif

namespace aggregpacking {


/*********************************************************************
* HETERO GEOMETRY CLASS
*********************************************************************/

void
HeteroGeometry::setAggregType(const string& s) {
	if (s!=Sphere::className && s!=Ellipsoid::className) {
		LOG_ERROR("unsupported aggregType values")
	}
	aggregType = s;
}

void
HeteroGeometry::build(const AggregSizeRandomizer& aggregSizeRandomizer, const ConcreteMixture& concreteMixture, double minRelAggregDist) {
	// do nothing if already built (from previous build or from load)
	if (aggregates.size() > 0) return;
	//
	take(aggregSizeRandomizer,concreteMixture);
	bool ok;
	ok = place(minRelAggregDist);
	if (!ok) { LOG_ERROR("place algorithm failed") }
#ifdef AGGREGPACKING_SERIALIZATION	
	save();
#endif
}

void
HeteroGeometry::rebuild(const AggregSizeRandomizer& aggregSizeRandomizer, const ConcreteMixture& concreteMixture, double minRelAggregDist) {
	delete octree;
	initOctree();
	clearAggregates();
	build(aggregSizeRandomizer,concreteMixture,minRelAggregDist);
}

void
HeteroGeometry::clearAggregates() {
	unsigned int n = aggregates.size();
	for (auto a : aggregates) delete a;
	aggregates.resize(0);
}

Aggregate*
HeteroGeometry::CreateParticle(double size) {
	Aggregate* a;
	if (aggregType == Sphere::className) {
		a = new Sphere();
	} else if (aggregType == Ellipsoid::className) {
		a = new Ellipsoid();
	} else {
		LOG_DEBUG("Unsupported aggregate type")
	}
	a->fromSize(size);
	return a;
}

void
HeteroGeometry::take(const AggregSizeRandomizer& aggregSizeRandomizer, const ConcreteMixture& concreteMixture) {
	double v = computeVolume();
	double temp = aggregSizeRandomizer.cdf(aggregSizeRandomizer.getMinSize());
	// actual already placed volume
	double aggrVolume=0;
	// actual volume to fill
	double finalAggrVolume = v*concreteMixture.aggregateVolume*(1.-temp);
	clearAggregates();
	Aggregate* a;
	double s;
	while (aggrVolume < finalAggrVolume) {
		// ranomly take a size
		s = aggregSizeRandomizer.take();
		// create a particle form this size
		a = CreateParticle(s);
		aggrVolume += a->computeVolume();
		aggregates.push_back(a);
	}
	// sort aggregates from the largest one to the smallest one
	std::sort(aggregates.begin(), aggregates.end(), [](Aggregate* a1, Aggregate* a2){ return a2->getSize() < a1->getSize(); } );
	int i = 0;
	for (auto a : aggregates) a->setNumber(i++);
}

bool
HeteroGeometry::place(double minRelAggregDist) {
	unsigned int i, j, k, warnInterval=1000;
	double d;
	bool posNotOK;
	bool tooBigOverlap;
	bool isOutside;
	bool prints;
	const double scale=1.5;
	const double shiftCoeff=(1.+minRelAggregDist)*1.1;
	double minDist;
	vector<Aggregate*> neighbors;
	Aggregate *a, *b;
	Vector3 normal,pos;
	for (auto a : aggregates) { // for all aggregates...
		posNotOK = true;
		i = 0;
		prints = false;
		minDist = a->getSize()*minRelAggregDist;
		while (posNotOK) { // while position is not ok
			a->setOri(Quaternion(Randomizer::random(),Randomizer::random(),Randomizer::random(),Randomizer::random()));
			posNotOK = false;
			isOutside = false;
			if (!placeOneParticle(*a)) {
				LOG_WARN("Particle could not be initially placed")
				return false;
			}
			neighbors.clear();
			octree->getPotentialNeighbors(neighbors,a,scale);
			for (k=0; k<10; k++) { // try to change position of particle and check neighbors again. Try only limited tries
				posNotOK = false;
				tooBigOverlap = false;
				j = 0;
				for (auto b : neighbors) { // check all neighbors
					j++;
					posNotOK = false;
					d = a->computeDistanceFrom(*b);
					if (d < minDist) { // not acceptable
						//std::cout<<d<< " "<<minDist<<"\n";
						if (-d > 2*a->getSize()) tooBigOverlap = true; // change position immediayelly if overlap is too big
						else { // else change position not to overlap currently overlapping particle
							pos = a->getPos();
							normal = pos - b->getPos();
							normal.normalize();
							pos += shiftCoeff*normal;
							a->setPos(pos);
						}
						if (!containsParticle(*a)) isOutside = true; // checks if the shift does not move the particle outside
						posNotOK = true;
						break; // break and try another position relatively close
					}
					if (tooBigOverlap || isOutside) break; // breaks under too bad conditions
				} // now all nigbors are ok
				if (!posNotOK || isOutside) break; // pos is ok. Or if particle is outside, do not continue
			}
			if (i == warnInterval) { // some info each warnInterval
				std::cerr << "HeteroGeometry: Aggregate size " << a->getSize() << " tries: " << warnInterval;
				prints = true;
			} else if (i !=0 && i != warnInterval && i%warnInterval == 0) { // some info each warnInterval
				std::cerr << " " << i;
			} else if (i > placeNumTries) { // after certain unsuccesful tries, throw error
				LOG_WARN("Particle could not be placed");
				return false;
			}
			i++;
		} // now position of particle is ok
		if (prints) std::cerr << " OK\n";
		octree->addData(a);
	}
	return true;
}

bool
HeteroGeometry::check() {
	if (aggregates.size() == 0) return true;
	if (!containsParticle(*(aggregates[0]))) return false;
	Aggregate *a, *b;
	unsigned int i,j;
	for (i=1; i<aggregates.size(); i++) {
		a = aggregates[i];
		if (!containsParticle(*a)) return false;
		for (j=0; j<i; j++) {
			b = aggregates[j];
			if (a->computeDistanceFrom(*b) < 0) return false;
		}
	}
	return true;
}

#ifdef AGGREGPACKING_SERIALIZATION
/// HeteroGeometry to ostream shift operator
ostream& operator<<(ostream& os, const HeteroGeometry& o) {
	const CubeHeteroGeometry* chg = dynamic_cast<const CubeHeteroGeometry*>(&o);
	if (chg) {
		os << *chg;
		return os;
	}
	LOG_ERROR("unsupported argument type")
}

/// HeteroGeometry from istream shift operator
istream& operator>>(istream& is, HeteroGeometry& o) {
	CubeHeteroGeometry* chg = dynamic_cast<CubeHeteroGeometry*>(&o);
	if (chg) {
		is >> *chg;
		return is;
	}
	LOG_ERROR("unsupported argument type")
}

void
HeteroGeometry::saveAggregates(ostream& os) const {
	unsigned int n = aggregates.size();
	os << n << "\n";
	for (auto a : aggregates) os << *a << "\n";
}

void
HeteroGeometry::loadAggregates(istream& is) {
	clearAggregates();
	unsigned int n;
	is >> n;
	aggregates.resize(n);
	Aggregate* a;
	for (unsigned int i=0; i<n; i++) {
		a = Aggregate::Load(is);
		aggregates[i] = a;
	}
}

#endif




/*********************************************************************
* CUBE HETERO GEOMETRY CLASS
*********************************************************************/
#ifdef AGGREGPACKING_SERIALIZATION	
CubeHeteroGeometry::CubeHeteroGeometry(double c, const string& m) : HeteroGeometry(m), cubeSize(c) {
	initOctree();
	load();
}
#else
CubeHeteroGeometry::CubeHeteroGeometry(double c) : HeteroGeometry(), cubeSize(c) {
	initOctree();
}
#endif

CubeHeteroGeometry::~CubeHeteroGeometry() {
	clearAggregates();
	delete octree;
}

void
CubeHeteroGeometry::initOctree() {
	octree = new Octree<Aggregate*>(.5*cubeSize,.5*cubeSize,.5*cubeSize,cubeSize,5);
}

bool
CubeHeteroGeometry::placeOneParticle(Aggregate& aggreg) {
	bool b = true;
	int i = 0;
	while (b) {
		double x = cubeSize * Randomizer::random();
		double y = cubeSize * Randomizer::random();
		double z = cubeSize * Randomizer::random();
		aggreg.setPos(x,y,z);
		b = !(aggreg.isInsideCube(cubeSize));
		if (i++ >= 30) { return false; }
	}
	return true;
}

bool
CubeHeteroGeometry::containsParticle(const Aggregate& aggreg) {
	return aggreg.isInsideCube(cubeSize);
}

#ifdef AGGREGPACKING_SERIALIZATION
const string CubeHeteroGeometry::className = "CubeHeteroGeometry";

/// CubeHeteroGeometry to ostream shift operator
ostream& operator<<(ostream& os, const CubeHeteroGeometry& o) {
	os << CubeHeteroGeometry::className << "\n" << o.cubeSize << "\n";
	o.saveAggregates(os);
	return os;
}

/// CubeHeteroGeometry from istream shift operator
istream& operator>>(istream& is, CubeHeteroGeometry& o) {
	double& c = o.cubeSize;
	is >> c;
	o.loadAggregates(is);
	//
	delete o.octree;
	o.octree = new Octree<Aggregate*>(.5*c,.5*c,.5*c,c,5);
	for (auto a : o.aggregates) o.octree->addData(a);
	return is;
}
#endif








/*********************************************************************
* PYTHON
*********************************************************************/
#ifdef AGGREGPACKING_BOOST_PYTHON
/// @cond DOXYGEN_EXCLUDE
void py_heterogeometry() {
	using namespace boost::python;

	class_<HeteroGeometry,boost::noncopyable>("HeteroGeometry",no_init);

	double (CubeHeteroGeometry::*CHGgetCubeSize)() = &CubeHeteroGeometry::getCubeSize;
	Octree<Aggregate*>* (CubeHeteroGeometry::*CHGgetOctree)() = &CubeHeteroGeometry::getOctree;
	vector<Aggregate*>& (CubeHeteroGeometry::*CHGgetAggregates)() = &CubeHeteroGeometry::getAggregates;
	string (CubeHeteroGeometry::*CHGgetAggregType)() = &CubeHeteroGeometry::getAggregType;
	void (CubeHeteroGeometry::*CHGbuild1)(const AggregSizeRandomizer&,const ConcreteMixture&,double) = &CubeHeteroGeometry::build;
	void (CubeHeteroGeometry::*CHGbuild2)(const AggregSizeRandomizer&,const ConcreteMixture&) = &CubeHeteroGeometry::build;
#ifdef AGGREGPACKING_SERIALIZATION
	void (CubeHeteroGeometry::*CHGsave1)() const = &CubeHeteroGeometry::save;
	void (CubeHeteroGeometry::*CHGsave2)(const string&) const = &CubeHeteroGeometry::save;
	bool (CubeHeteroGeometry::*CHGload1)() = &CubeHeteroGeometry::load;
	bool (CubeHeteroGeometry::*CHGload2)(const string&) = &CubeHeteroGeometry::load;
#endif
	class_<CubeHeteroGeometry,boost::noncopyable>("CubeHeteroGeometry",init<double>())
#ifdef AGGREGPACKING_SERIALIZATION
		.def(init<double,const string&>())
#endif
		.def("computeVolume",&HeteroGeometry::computeVolume)
		.add_property("aggregates",make_function(CHGgetAggregates,return_internal_reference<>()))
		.add_property("volume",&HeteroGeometry::computeVolume)
		.add_property("cubeSize",CHGgetCubeSize)
		.add_property("octree",make_function(CHGgetOctree,return_internal_reference<>()))
		.add_property("aggregType",CHGgetAggregType,&HeteroGeometry::setAggregType)
		.add_property("fileName",&HeteroGeometry::getFileName)
		.add_property("placeNumTries",&HeteroGeometry::getPlaceNumTries,&HeteroGeometry::setPlaceNumTries)
		.def("check",&HeteroGeometry::check)
		.def("build",CHGbuild1)
		.def("build",CHGbuild2)
#ifdef AGGREGPACKING_SERIALIZATION
		.def("save",CHGsave1)
		.def("save",CHGsave2)
		.def("load",CHGload1)
		.def("load",CHGload2)
#endif
		;
}
/// @endcond
#endif

} // namespace
