/*
*
*      AggregPacking : Generator of random aggregate packings in concrete     
*                                                                             
*                           version 0.1 (2014-08-22)                          
* 
*                      Copyright (C)  2014  Jan Stransky                      
* 
*          Czech Technical University, Faculty of Civil Engineering,          
*      Department of Structural Mechanics, 166 29 Prague, Czech Republic      
* 
* AggregPacking is free software: you can redistribute it and/or modify it
* under the terms of the GNU Lesser General Public License as published by the
* Free Software Foundation, either version 3 of the License, or (at your
* option) any later version.
* 
* AggregPacking is distributed in the hope that it will be useful, but WITHOUT
* ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
* FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License
* for more details.
* 
* You should have received a copy of the GNU Lesser General Public License
* along with this program. If not, see <http://www.gnu.org/licenses/>.
*/

/**
 * @file
 * @brief File containing representation of actual heterogeneous geometry
 */

#ifndef HETEROGEOMETRY_HPP
#define HETEROGEOMETRY_HPP

#include "utils.hpp"
#include "aggregate.hpp"
#include "randomizer.hpp"
#include "concmixture.hpp"
#include "octree.hpp"
using octree::Octree;

#include <string>
using std::string;

#ifdef AGGREGPACKING_SERIALIZATION
#include <iostream>
#include <fstream>
using std::istream;
using std::ostream;
using std::ifstream;
using std::ofstream;
#endif

namespace aggregpacking {


/*********************************************************************
* HETERO GEOMETRY
*********************************************************************/
/// Abstract class representing heterogeneous geometry
/**
 * Contains a list of aggregates and functions for generation
 */
class HeteroGeometry {
	protected:
#ifdef AGGREGPACKING_SERIALIZATION
		/// Default constructor
		HeteroGeometry() : HeteroGeometry(string()) {};

		/// Default constructor from file name
		/** @param fileName file name
		 */
		HeteroGeometry(const string& fileName) : fileName(fileName), aggregType(Sphere::className), aggregates(0), placeNumTries(10000) {};
#else
		HeteroGeometry() : aggregType(Sphere::className), aggregates(0), placeNumTries(10000) {};
#endif

		/// Aggregates container
		vector<Aggregate*> aggregates;

		/// File name for saving files
		string fileName;

		/// Octree (for faster spatial checks)
		Octree<Aggregate*>* octree;

		/// How many tries of one particle placement is done until error is thrown (default 10000)
		int placeNumTries;

		/// Creates and return new Aggregate particle according to aggregType and given size
		/** @param size size of returned particle
		 * @return new Aggregate object
		 */
		Aggregate* CreateParticle(double size);

		/// Creates a random particle
		/** @param aggregSizeRandomizer sieve curve random size generator
		 *  @param concreteMixture used concrete mixture
		 */
		void take(const AggregSizeRandomizer& aggregSizeRandomizer, const ConcreteMixture& concreteMixture);

		/// Places all particles randomly such that they have guaranteed distance between each other and all are entirely contained by the cube
		/** @param minRelAggregDist minimal distance between particles, relative to the smaller particle from each pair. I.e., the distance is at least (1.+minRelAggregDist)*sizeOfSmallerParticle
		 * @return true if successful, false otherwise
		 */
		bool place(double minRelAggregDist);

		/// Places all particles randomly such that they do not overlap and all are entirely contained by the cube
		/** @return true if successful, false otherwise
		 */
		bool place() { return place(0.); }

		/// Places one particle such that it does not overlap with already placed particles and is entirely contained by the cube
		/** @param aggreg given particle
		 * @return true if succesful, false otherwise
		 */
		virtual bool placeOneParticle(Aggregate& aggreg) = 0;

		/// Tests if given particle is entirely contained by receiver
		/** @param aggreg given particle
		 * @return true if receiver contains entire given particle, false otherwise
		 */
		virtual bool containsParticle(const Aggregate& aggreg) = 0;

		/// Delete all aggregates
		void clearAggregates();

		/// Aggregate type to generate. Currently implemented are "Sphere" and "Ellipsoid"
		string aggregType;

#ifdef AGGREGPACKING_SERIALIZATION
		/// Saves receiver to given ostream
		void saveAggregates(ostream& os) const;

		/// Loads receiver from given istream
		void loadAggregates(istream& is);
#endif
	public:
		/// Generate the geometry by take-and-place process. Aggregates has at least guaranteed distance between each other. Do nothing if already built
		/** @param aggregSizeRandomizer sieve curve random size generator
		 *  @param concreteMixture used concrete mixture
		 *  @param minRelAggregDist minimum relative aggregates distance (relatively to the smaller particle)
		 */
		void build(const AggregSizeRandomizer& aggregSizeRandomizer, const ConcreteMixture& concreteMixture, double minRelAggregDist);

		/// Generate the geometry by take-and-place process. Aggregates do not overlap. Do nothing if already built
		/** @param aggregSizeRandomizer sieve curve random size generator
		 *  @param concreteMixture used concrete mixture
		 */
		void build(const AggregSizeRandomizer& aggregSizeRandomizer, const ConcreteMixture& concreteMixture) { build(aggregSizeRandomizer,concreteMixture,0.); }

		/// Delete already build and re-generate the geometry by take-and-place process. Aggregates has at least guaranteed distance between each other
		/** @param aggregSizeRandomizer sieve curve random size generator
		 *  @param concreteMixture used concrete mixture
		 *  @param minRelAggregDist minimum relative aggregates distance (relatively to the smaller particle)
		 */
		void rebuild(const AggregSizeRandomizer& aggregSizeRandomizer, const ConcreteMixture& concreteMixture, double minRelAggregDist);

		/// Delete already build and re-generate the geometry by take-and-place process. Aggregates do not overlap
		/** @param aggregSizeRandomizer sieve curve random size generator
		 *  @param concreteMixture used concrete mixture
		 */
		void rebuild(const AggregSizeRandomizer& aggregSizeRandomizer, const ConcreteMixture& concreteMixture) { rebuild(aggregSizeRandomizer,concreteMixture,0.); }

		/// Compute volume of receiver
		virtual double computeVolume() = 0;

		/// Initialize new octree (and delete the old one)
		virtual void initOctree() = 0;

		/// Getter for aggregates (return by const reference)
		const vector<Aggregate*>& getAggregates() const { return aggregates; }

		/// Getter for aggregates (return by non-const reference)
		vector<Aggregate*>& getAggregates() { return aggregates; }

		/// Getter for aggregate type (by value)
		string getAggregType() { return aggregType; }

		/// Getter for aggregate type (by const reference)
		const string& getAggregType() const { return aggregType; }

		/// Setter for aggreg type
		void setAggregType(const string& s);

		/// Getter for octreee (by non-const pointer)
		Octree<Aggregate*>* getOctree() { return octree; }

		/// Getter for octreee (by const pointer)
		const Octree<Aggregate*>* getOctree() const { return octree; }

		/// Getter for file name
		string getFileName() { return fileName; }

		/// Getter for placeNumTries
		int getPlaceNumTries() { return placeNumTries; }

		/// Setter for placeNumTries
		void setPlaceNumTries(int v) { placeNumTries = v; }

		/// Checks that no particles overlap
		/** @return true if no particles overlap, false if any do
		 */
		bool check();

#ifdef AGGREGPACKING_SERIALIZATION
		/// Returns string representation of Sphere class name
		virtual const string& getClassName() = 0;

		/// HeteroGeometry to ostream shift operator
		friend ostream& operator<<(ostream& f, const HeteroGeometry& o);

		/// HeteroGeometry from istream shift operator
		friend istream& operator>>(istream& f, HeteroGeometry& o);

		/// Saves receiver to given file
		/** @param fName file name
		 */
		void save(const string& fName) const { saveObject(*this,fName,fileName); }

		/// Saves receiver to this->fileName
		void save() const { save(string()); }

		/// Loads receiver from given file
		/** @param fName file name
		 */
		bool load(const string& fName) { return loadObject(*this,fName,fileName); }

		/// Saves receiver from this->fileName
		bool load() { return load(string()); }
#endif
};






/*********************************************************************
* CUBE HETERO GEOMETRY
*********************************************************************/
/// Abstract class representing heterogeneous geometry
/**
 * Contains a list of aggregates and functions for generation
 */
class CubeHeteroGeometry : public HeteroGeometry {
	private:
		/// Cube size of receiver
		double cubeSize;

		virtual bool placeOneParticle(Aggregate& aggreg);

	public:
#ifdef AGGREGPACKING_SERIALIZATION
		/// Constructor from size and file name
		/** @param c cube size
		 *  @param m file name
		 */
		CubeHeteroGeometry(double c, const string& m);

		/// Constructor from size
		/** @param c cube size
		 */
		CubeHeteroGeometry(double c) : CubeHeteroGeometry(c,string()) {}
#else
		CubeHeteroGeometry(double c);
#endif

		/// Destructor
		~CubeHeteroGeometry();

		virtual void initOctree();

		/// Compute volume of cube
		virtual double computeVolume() { return pow(cubeSize,3); }

		/// Getter for cube size
		double getCubeSize() { return cubeSize; }

		virtual bool containsParticle(const Aggregate& aggreg);

#ifdef AGGREGPACKING_SERIALIZATION
		/// String representation of CubeHeteroGeometry class name
		const static string className;

		const string& getClassName() { return CubeHeteroGeometry::className; }

		/// CubeHeteroGeometry to ostream shift operator
		friend ostream& operator<<(ostream& f, const CubeHeteroGeometry& o);

		/// CubeHeteroGeometry from istream shift operator
		friend istream& operator>>(istream& f, CubeHeteroGeometry& o);
#endif
};





/*********************************************************************
* PYTHON
*********************************************************************/
#ifdef AGGREGPACKING_BOOST_PYTHON
/// @cond DOXYGEN_EXCLUDE
void py_heterogeometry();
/// @endcond
#endif

} // namespace
#endif
