/*
*
*      AggregPacking : Generator of random aggregate packings in concrete     
*                                                                             
*                           version 0.1 (2014-08-22)                          
* 
*                      Copyright (C)  2014  Jan Stransky                      
* 
*          Czech Technical University, Faculty of Civil Engineering,          
*      Department of Structural Mechanics, 166 29 Prague, Czech Republic      
* 
* AggregPacking is free software: you can redistribute it and/or modify it
* under the terms of the GNU Lesser General Public License as published by the
* Free Software Foundation, either version 3 of the License, or (at your
* option) any later version.
* 
* AggregPacking is distributed in the hope that it will be useful, but WITHOUT
* ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
* FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License
* for more details.
* 
* You should have received a copy of the GNU Lesser General Public License
* along with this program. If not, see <http://www.gnu.org/licenses/>.
*/

/**
 * @file
 * @brief File containing octree representation
 */

#ifndef OCTREE_HPP
#define OCTREE_HPP

#ifdef AGGREGPACKING_BOOST_PYTHON
#include <boost/python.hpp>
#endif

#include <vector>
#include <iterator>
using std::vector;

/// Namespace octree
namespace octree {


template<class T> class OctreeNode;

/**********************************************************************
* OCTREE ITERATOR
**********************************************************************/
// TODO reduce code duplication
/// Octree iterator
template<class T>
class OctreeIterator : public std::iterator<std::input_iterator_tag, OctreeNode<T>> {
	protected:
		/// stored octree node
		OctreeNode<T>* ptr;

	public:
		/// Constructor from octree node pointer
		OctreeIterator(OctreeNode<T>* ptr) : ptr(ptr) {}

		/// Copy constructor
		OctreeIterator(const OctreeIterator& it) : ptr(it.ptr) {}

		/// Operator++
		OctreeIterator& operator++();

		/// ++operator
		OctreeIterator operator++(int) {OctreeIterator tmp(*this); operator++(); return tmp; }

		/// Equality operator
		bool operator==(const OctreeIterator& rhs) { return ptr == rhs.ptr; }

		/// Inequality operator
		bool operator!=(const OctreeIterator& rhs) { return ptr != rhs.ptr; }

		/// Dereference operator
		OctreeNode<T>& operator*() { return *(this->ptr); }

		/// Operator ->
		OctreeNode<T>* operator->() { return this->ptr; }
};

/// Octree iterator
template<class T>
class OctreeConstIterator : public std::iterator<std::input_iterator_tag, OctreeNode<T>> {
	protected:
		/// stored octree node
		OctreeNode<T>* ptr;

	public:
		/// Constructor from octree node pointer
		OctreeConstIterator(OctreeNode<T>* ptr) : ptr(ptr) {}

		/// Copy constructor
		OctreeConstIterator(const OctreeConstIterator& it) : ptr(it.ptr) {}

		/// Operator++
		OctreeConstIterator& operator++();

		/// ++operator
		OctreeConstIterator operator++(int) {OctreeConstIterator tmp(*this); operator++(); return tmp; }

		/// Equality operator
		bool operator==(const OctreeConstIterator& rhs) { return ptr == rhs.ptr; }

		/// Inequality operator
		bool operator!=(const OctreeConstIterator& rhs) { return ptr != rhs.ptr; }

		/// Dereference operator
		const OctreeNode<T>& operator*() const { return *(this->ptr); }

		/// Operator ->
		const OctreeNode<T>* operator->() const { return this->ptr; }
};






/**********************************************************************
* OCTREE
**********************************************************************/
/// Class representing full octree
template <class T>
class Octree {
	private:
		/// Depth (number of levels)
		int depth;

		/// Root node
		OctreeNode<T>* root;

	public:
		/// Default constructor
		Octree() : depth(0), root() {}

		/// Constructor
		/** @param x x coordinate of center
		 *  @param y y coordinate of center
		 *  @param z z coordinate of center
		 *  @param size dimension of octree
		 *  @param depth final delth (number of levels)
		 */
		Octree(double x, double y, double z, double size, int depth);

		/// Constructor
		/** @param x x coordinate of center
		 *  @param y y coordinate of center
		 *  @param z z coordinate of center
		 *  @param size dimension of octree
		 */
		Octree(double x, double y, double z, double size) : Octree(x,y,z,size,0) {};

		/// Destructor
		~Octree();

		/// Getter for depth
		int getDepth() { return depth; }

		/// Getter for size
		double getSize() { return root->getSize(); }

		/// Builds the octree structure
		/** @param x x coordinate of the ceter
		 *  @param y y coordinate of the ceter
		 *  @param z z coordinate of the ceter
		 *  @param size dimension of octree
		 */
		void build(double x, double y, double z, double size);

		/// Getter for root (by pointer)
		OctreeNode<T>* getRoot() { return root; }

		/// Getter for root (by const pointer)
		const OctreeNode<T>* getRoot() const { return root; }

		/// Adds data to Octree
		void addData(const T& data);

		/// Returns total number of nodes of receiver
		int getNumberOfNodes();

		/// Octree iterator
		typedef OctreeIterator<T> iterator; // for boost python

		/// Octree iterator begin
		OctreeIterator<T> begin() { return OctreeIterator<T>(root); }

		/// Octree iterator end
		OctreeIterator<T> end() { return OctreeIterator<T>(nullptr); }

		/// Octree const iterator begin
		OctreeConstIterator<T> begin() const { return OctreeConstIterator<T>(root); }

		/// Octree const iterator end
		OctreeConstIterator<T> end() const { return OctreeConstIterator<T>(nullptr); }

		/// Gets potential neighbors around given point and size
		/** @param answer resulting vector of neighbor objects
		 *  @param x x coordinate of center of object
		 *  @param y y coordinate of center of object
		 *  @param z z coordinate of center of object
		 *  @param size size of object
		 */
		void getPotentialNeighbors(vector<T>& answer, double x, double y, double z, double size) const;

		/// Gets potential neighbors for given object
		/** @param answer resulting vector of neighbor objects
		 *  @param o given object
		 */
		void getPotentialNeighbors(vector<T>& answer, const T& o) const { getPotentialNeighbors(answer,get_x(o), get_y(o), get_z(o), get_size(o)); }

		/// Gets potential neighbors for given object, considering its scaled size
		/** @param answer resulting vector of neighbor objects
		 *  @param o given object
		 *  @param scale scale for object's size
		 */
		void getPotentialNeighbors(vector<T>& answer, const T& o, int scale) const { getPotentialNeighbors(answer,get_x(o), get_y(o), get_z(o), get_size(o)*scale); }

#ifdef AGGREGPACKING_BOOST_PYTHON
		/// getPotentialNeighbors function for Python
		vector<T>* py_getPotentialNeighbors(double x, double y, double z, double size) const;

		/// getPotentialNeighbors function for Python
		vector<T>* py_getPotentialNeighbors(const T& o) const { return py_getPotentialNeighbors(get_x(o), get_y(o), get_z(o), get_size(o)); }

		/// getPotentialNeighbors function for Python
		vector<T>* py_getPotentialNeighbors(const T& o,int scale) const { return py_getPotentialNeighbors(get_x(o), get_y(o), get_z(o), get_size(o)*scale); }
#endif
};




/**********************************************************************
* OCTREE
**********************************************************************/
/// Class representing octree node
template <class T>
class OctreeNode {
	private:
		/// x coordinate
		double x;

		/// y coordinate
		double y;

		/// z coordinate
		double z;

		/// Child nodes
		OctreeNode** children;

		/// Vector of data belonging to the node
		vector<T> data;

		/// Parent node
		OctreeNode* parent;

		/// level of the node
		int level;

		/// Dimension of the node
		double size;
		
		/// Number of the node
		int number;
		
		/// Constructor, used by a parent node
		/** @param x x coordinate
		 *  @param y y coordinate
		 *  @param z z coordinate
		 *  @param parent parent node
		 *  @param number number of the node
		 */
		OctreeNode(double x, double y, double z, OctreeNode& parent, int number);

		/// Constructor to create root node
		/** @param x x coordinate
		 *  @param y y coordinate
		 *  @param z z coordinate
		 *  @param size size of the node
		 */
		OctreeNode(double x, double y, double z, double size);

		/// Divides receiver until defined depth
		/** @param depth defined depth (number of levels) of the division
		 */
		void divide(int depth);

	public:
		/// Destructor
		~OctreeNode();

		/// Equality operator, intended for Python
		bool operator==(const OctreeNode& other) { return &other == this; }

		/// Inequality operator, intended for Python
		bool operator!=(const OctreeNode& other) { return &other != this; }

		/// Getter for x coordinate 
		double getX() { return x; }

		/// Getter for y coordinate 
		double getY() { return y; }

		/// Getter for z coordinate 
		double getZ() { return z; }

		/// Getter for size
		double getSize() { return size; }

		/// Getter for level
		int getLevel() { return level; }

		/// Getter for number
		int getNumber() { return number; }

		/// Getter for data (by const reference)
		const vector<T>& getData() const { return data; }

		/// Getter for data (by non-const reference)
		vector<T>& getData() { return data; }

		/// Getter for children (by const pointer to pointer)
		const OctreeNode** getChildren() const { return children; }

		/// Getter for children (as a vector, intended for Python)
		vector<OctreeNode*>* py_getChildren();

		/// Getter for parent (by pointer)
		OctreeNode* getParent() { return parent; }

		/// Getter for parent (by const pointer)
		const OctreeNode* getParent() const { return parent; }

		/// Returns if receiver is root or not
		bool isRoot() const { return parent == nullptr; }

		/// Returns if receiver is leaf or not
		bool isLeaf() const { return children == nullptr; }

		/// Adds data to receiver or its children. Determines x,y,z and size from given data
		void addData(const T& data);

		/// Adds data to receiver or its children. x,y,z and size are given
		void addData(const T& data, double x, double y, double z, double size);

		/// Returns if receiver contains entirely given box or not
		/** @param x x coordinate of box center
		 *  @param y y coordinate of box center
		 *  @param z z coordinate of box center
		 *  @param size size fo the box
		 */
		bool containsEntireBox(double x, double y, double z, double size) const;

		/// Returns if receiver contains entirely axis aligned box made around given object
		bool containsEntireBox(const T& o) const { return containsEntireBox(X_Y_Z_FROM_T(o)); }

		/// Returns children intersecting given box
		/** @param x x coordinate of box center
		 *  @param y y coordinate of box center
		 *  @param z z coordinate of box center
		 *  @param size size fo the box
		 *  @return bitmask. For example 0b00100101, means that 1st, 3rd and 6th children intersects box, the others not
		 */
		int getChildrenIntersectingBox(double x, double y, double z, double size) const;

		/// Returns children of receiver receiver intersecting axis aligned box made around given object
		int getChildrenIntersectingBox(const T& o) const { return getChildrenIntersectingBox(get_x(o), get_y(o), get_z(o), get_size(o)); }

#ifdef AGGREGPACKING_BOOST_PYTHON
		/// Returns vector of children intersecting box, intended for Python
		vector<OctreeNode*>* py_getChildrenIntersectingBox(double x, double y, double z, double size) const;

		/// Returns vector of children intersecting box, intended for Python
		vector<OctreeNode*>* py_getChildrenIntersectingBox(const T& o) const { return py_getChildrenIntersectingBox(get_x(o), get_y(o), get_z(o), get_size(o)); }
#endif

		/// Getter for root (by pointer)
		OctreeNode* getRoot();

		/// Getter for root (by const pointer)
		const OctreeNode* getRoot() const;

		/// Returns next node
		/** In this order:
		 *   - if receiver has children, return its first child
		 *   - if receiver is root and has no children, return nullptr
		 *   - if receiver is not the last child of its parent, return its next sibling
		 *   - find receiver's ancestor, which is not the last child of its parent. If such ancestor is root, return nullptr. Else return next sibling of this ancestor
		 */
		OctreeNode* next() const;

		/// returns vector of objects from receiver and its children
		/** @param answer resulting vetor of objects, potential neighbors
		 *  @param x x coordinate of given object
		 *  @param y y coordinate of given object
		 *  @param z z coordinate of given object
		 *  @param size size of given object
		 */
		void getPotentialNeighbors(vector<T>& answer, double x, double y, double z, double size) const;

		/// returns vector of objects from receiver, its neighbors and their children
		/** @param answer resulting vetor of objects, potential neighbors
		 *  @param o given object
		 */
		void getPotentialNeighbors(vector<T>& answer, const T& o) const { getPotentialNeighbors(answer,get_x(o), get_y(o), get_z(o), get_size(o)); }

#ifdef AGGREGPACKING_BOOST_PYTHON
		/// Returns new vector of potential neighbors, intended for Python
		vector<T>* py_getPotentialNeighbors(double x, double y, double z, double size) const;

		/// Returns new vector of potential neighbors, intended for Python
		vector<T>* py_getPotentialNeighbors(const T& o) const { return py_getPotentialNeighbors(get_x(o), get_y(o), get_z(o), get_size(o)); }
#endif

		/// Auxiliary function to extract indexed value from bitmask flag
		/** @param flag bitmask flag
		 *  @param index index to check
		 */
		static inline bool bitFlag2bool(int flag, int index) { return (bool)(flag & (1 << index) ); }

		// Auxiliary function // TODO
		//static inline bool quickIntersectionCheck(const T& o, double x, double y, double z, double size);

	/// Friend to enable Octree build
	friend void Octree<T>::build(double x, double y, double z, double size);
};

#include "octree.tcc"


} // namespace
#endif
