/*
*
*      AggregPacking : Generator of random aggregate packings in concrete     
*                                                                             
*                           version 0.1 (2014-08-22)                          
* 
*                      Copyright (C)  2014  Jan Stransky                      
* 
*          Czech Technical University, Faculty of Civil Engineering,          
*      Department of Structural Mechanics, 166 29 Prague, Czech Republic      
* 
* AggregPacking is free software: you can redistribute it and/or modify it
* under the terms of the GNU Lesser General Public License as published by the
* Free Software Foundation, either version 3 of the License, or (at your
* option) any later version.
* 
* AggregPacking is distributed in the hope that it will be useful, but WITHOUT
* ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
* FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License
* for more details.
* 
* You should have received a copy of the GNU Lesser General Public License
* along with this program. If not, see <http://www.gnu.org/licenses/>.
*/

#include "randomizer.hpp"
#include "utils.hpp"

#include <cmath>
#include <stdexcept>

#ifdef AGGREGPACKING_SERIALIZATION
#include <string>
using std::string;
#endif


namespace aggregpacking {


/**********************************************************************
* RANDOMIZER
**********************************************************************/
int Randomizer::seed = 0;



/**********************************************************************
* AGGREGATE SIZE RANDOMIZER
**********************************************************************/

void
AggregSizeRandomizer::initSizes(unsigned int nMinValues) {
#ifdef AGGREGPACKING_SERIALIZATION
	// try to load already saved sizes
	if (load()) return;
#endif
	double dmin = this->getMinSize(), dmax = this->getMaxSize();
	// choose initial delta
	double delta0 = (dmax-dmin)*.01;
	int ii = 0, jj, kk;
	double size, delta, n;
	bool finish;
	do { // do loop while length of sizes is large enough
		sizes.resize(1);
		sizes[0] = dmin;
		size = dmin;
		// make delta smaller than in previous try
		delta = delta0*pow(2,-ii);
		// compute relative number of particles in interval (dmin,dmin+delta)
		n = (cdf(size+delta)-cdf(size))/pow(size,3);
		jj = 0;
		while (size < dmax) { // fill sizes from dmin dmax dmax
			finish = false;
			double min = 0, max=dmax-size, vmin, vmax, vdelta;
			kk = 0;
			#define F(d) (cdf(size+d)-cdf(size))/pow(size,3) - n
			// by bisection method, find delta such that relative number in interval (size,size+delta) is equal to precomputed value n
			do {
				delta = .5*(min+max);
				vmin = F(min);
				vmax = F(max);
				if (vmin>0) { LOG_ERROR("Bisection method failed") }
				if (vmax<0) { // should only hapen at the end of interval
					// stop the method
					finish = true;
					break;
				}
				vdelta = F(delta);
				if (std::abs(vdelta) < 1e-12) break;
				else if (vdelta < 0) min = delta;
				else if (vdelta > 0) max = delta;
				else { LOG_ERROR("Bisection method failed") } // this is impossible to happen if I am right
				if (kk++ > 100) { LOG_ERROR("Bisection method did not converge") }
			} while (max/min-1 > 1e-9);
			#undef F
			if (finish) {
				sizes.push_back(getMaxSize());
				break;
			} else {
				size += delta;
				sizes.push_back(size);
			}
		}
		ii++;
	} while (sizes.size() < nMinValues);
#ifdef AGGREGPACKING_SERIALIZATION
	save();
#endif
}

double
AggregSizeRandomizer::take() const {
	// choose randomly one interval from sizes and choose randomly one number from that interval. Both randoms has uniform distribution
	const int i = randint(sizes.size()-2);
	const double& s1 = sizes[i], s2 = sizes[i+1];
	double ds = s2-s1;
	double r = random();
	double ret = s1 + ds*r;
	if (ret > getMaxSize()) {
		return take();
	}
	return ret;
}

#ifdef AGGREGPACKING_SERIALIZATION
void
AggregSizeRandomizer::saveSizes(ostream& os) const {
	int size = sizes.size();
	os << size;
	for (auto& s : sizes) os << "\n" << s;
	os << "\n";
}

void
AggregSizeRandomizer::loadSizes(istream& is) {
	int size;
	is >> size;
	sizes.resize(size);
	double d;
	for (auto& s : sizes) is >> s;
}

AggregSizeRandomizer*
AggregSizeRandomizer::Load(istream& is) {
	string str;
	is >> str;
	if (str == FullerCurveRandomizer::className) {
		FullerCurveRandomizer* ret = new FullerCurveRandomizer();
		is >> *ret;
		return ret;
	}
	if (str == ExperimentalSieveCurveRandomizer::className) {
		ExperimentalSieveCurveRandomizer* ret = new ExperimentalSieveCurveRandomizer();
		is >> *ret;
		return ret;
	}
	LOG_ERROR2("Wrong indentifier:",str)
}

/// AggregSizeRandomizer to ostream shift operator
ostream& operator<<(ostream& os, const AggregSizeRandomizer& o) {
	const FullerCurveRandomizer* fcr = dynamic_cast<const FullerCurveRandomizer*>(&o);
	if (fcr) {
		os << *fcr;
		return os;
	}
	const ExperimentalSieveCurveRandomizer* escr = dynamic_cast<const ExperimentalSieveCurveRandomizer*>(&o);
	if (escr) {
		os << *escr;
		return os;
	}
	LOG_ERROR("unsupported argument type")
}

/// AggregSizeRandomizer from istream shift operator
istream& operator>>(istream& is, AggregSizeRandomizer& o) {
	FullerCurveRandomizer* fcr = dynamic_cast<FullerCurveRandomizer*>(&o);
	if (fcr) {
		is >> *fcr;
		return is;
	}
	ExperimentalSieveCurveRandomizer* escr = dynamic_cast<ExperimentalSieveCurveRandomizer*>(&o);
	if (escr) {
		is >> *escr;
		return is;
	}
	LOG_ERROR("unsupported argument type")
}
#endif



/**********************************************************************
* FULLER CURVE RANDOMIZER
**********************************************************************/
double 
FullerCurveRandomizer::cdf(double size) const {
	return pow(size/getMaxSize(),exponent);
}

#ifdef AGGREGPACKING_SERIALIZATION
const string FullerCurveRandomizer::className = "FullerCurveRandomizer";

/// FullerCurveRandomizer to ostream shift operator
ostream& operator<<(ostream& os, const FullerCurveRandomizer& o) {
	os << FullerCurveRandomizer::className << "\n" << o.exponent << "\n";
	o.saveSizes(os);
	return os;
}

/// FullerCurveRandomizer from istream shift operator
istream& operator>>(istream& is, FullerCurveRandomizer& o) {
	is >> o.exponent;
	o.loadSizes(is);
	return is;
}
#endif







/**********************************************************************
* EXPERIMENTAL SIEVE CURVE RANDOMIZER
**********************************************************************/
void
ExperimentalSieveCurveRandomizer::check() {
	double size=0., ammount=0.;
	for (int i=0; i<values.size(); i++) {
		if (values[i][0] <= size) {
			LOG_ERROR("values not increasing");
		}
		size = values[i][0];
		if (values[i][1] < ammount) {
			LOG_ERROR("values (ammounts) not non-decreasing");
		}
		ammount = values[i][1];
	}
	if (values[values.size()-1][1] != 1.) {
		LOG_ERROR("last ammount should be 1.0");
	}
	// first ammount can be arbitrary
}

int
ExperimentalSieveCurveRandomizer::findInterval(double size) const {
	size = std::max(size,getMinSize());
	for (int i=0; i<values.size()-1; i++) {
		if (values[i+1][0] > size) return i;
	}
	return values.size()-1;
}

double 
ExperimentalSieveCurveRandomizer::cdf(double size) const {
	// piecewise linear curve in semi-log scale -> piecewise log curve in real scale
	/* y0 = a*log(x0) + b; y1 = a*log(x1) + b;
	   b = y0 - a*log(x0); y1 = a*log(x1) - a*log(x0) + y0; y1-y0 = a*(log(x1)-log(x0)); a = (y1-y0)/(log(x1)-log(x0))
		b = y0 - a*log(x0);
	*/
	const int i = findInterval(size);
	const double& x0=values[i][0],x1=values[i+1][0],y0=values[i][1],y1=values[i+1][1];
	const double a = (y1-y0)/(log(x1)-log(x0));
	const double b = y0 - a*log(x0);
	return a*log(size) + b;
}

#ifdef AGGREGPACKING_SERIALIZATION
const string ExperimentalSieveCurveRandomizer::className = "ExperimentalSieveCurveRandomizer";

/// ExperimentalSieveCurveRandomizer to ostream shift operator
ostream& operator<<(ostream& os, const ExperimentalSieveCurveRandomizer& o) {
	os << ExperimentalSieveCurveRandomizer::className << "\n";
	const ExperimentalSieveCurveRandomizer::sieve_curve& ov = o.values;
	os << ov.size() << "\n";
	for (int i=0; i<ov.size(); i++) os << ov[i][0] << " " << ov[i][1] << "\n";
	o.saveSizes(os);
	return os;
}

/// ExperimentalSieveCurveRandomizer from istream shift operator
istream& operator>>(istream& is, ExperimentalSieveCurveRandomizer& o) {
	int temp;
	is >> temp;
	ExperimentalSieveCurveRandomizer::sieve_curve& ov = o.values;
	ov.resize(temp);
	for (int i=0; i<ov.size(); i++) is >> ov[i][0] >> ov[i][1];
	o.loadSizes(is);
	return is;
}
#endif






/**********************************************************************
* PYTHON
**********************************************************************/
#ifdef AGGREGPACKING_BOOST_PYTHON
/// @cond DOXYGEN_EXCLUDE
void py_randomizer() {
	using namespace boost::python;

	class_<Randomizer,boost::noncopyable>("Randomizer",no_init)
		.def("random",&Randomizer::random)
		.staticmethod("random")
		.def("randint",&Randomizer::randint)
		.staticmethod("randint")
		.add_static_property("seed",&Randomizer::getSeed,&Randomizer::setSeed)
		;

	class_<AggregSizeRandomizer,boost::noncopyable>("AggregSizeRandomizer",no_init);

#ifdef AGGREGPACKING_SERIALIZATION
	void (FullerCurveRandomizer::*FCRsave1)() const = &FullerCurveRandomizer::save;
	void (FullerCurveRandomizer::*FCRsave2)(const string&) const = &FullerCurveRandomizer::save;
	bool (FullerCurveRandomizer::*FCRload1)() = &FullerCurveRandomizer::load;
	bool (FullerCurveRandomizer::*FCRload2)(const string&) = &FullerCurveRandomizer::load;
#endif
	double (FullerCurveRandomizer::*FCRgetExponent)() = &FullerCurveRandomizer::getExponent;
	class_<FullerCurveRandomizer,bases<AggregSizeRandomizer>,boost::noncopyable>("FullerCurveRandomizer",init<>())
		.def(init<int, double, double, double>())
#ifdef AGGREGPACKING_SERIALIZATION
		.def(init<int, double, double, double, const string&>())
#endif
		.def("cdf",&FullerCurveRandomizer::cdf)
		.def("take",&FullerCurveRandomizer::take)
		.add_property("sizes",make_function(&FullerCurveRandomizer::getSizes,return_internal_reference<>()))
		.add_property("exponent",FCRgetExponent,&FullerCurveRandomizer::setExponent)
		.add_property("minSize",&FullerCurveRandomizer::getMinSize)
		.add_property("maxSize",&FullerCurveRandomizer::getMaxSize)
		.add_static_property("seed",&Randomizer::getSeed,&Randomizer::setSeed)
#ifdef AGGREGPACKING_SERIALIZATION
		.add_property("fileName",&FullerCurveRandomizer::getFileName)
		.def("save",FCRsave1)
		.def("save",FCRsave2)
		.def("load",FCRload1)
		.def("load",FCRload2)
#endif
		;

#ifdef AGGREGPACKING_SERIALIZATION
	void (ExperimentalSieveCurveRandomizer::*ESCRsave1)() const = &ExperimentalSieveCurveRandomizer::save;
	void (ExperimentalSieveCurveRandomizer::*ESCRsave2)(const string&) const = &ExperimentalSieveCurveRandomizer::save;
	bool (ExperimentalSieveCurveRandomizer::*ESCRload1)() = &ExperimentalSieveCurveRandomizer::load;
	bool (ExperimentalSieveCurveRandomizer::*ESCRload2)(const string&) = &ExperimentalSieveCurveRandomizer::load;
#endif
	ExperimentalSieveCurveRandomizer::sieve_curve (ExperimentalSieveCurveRandomizer::*ESCRgetValues)() = &ExperimentalSieveCurveRandomizer::getValues;
	class_<ExperimentalSieveCurveRandomizer,bases<AggregSizeRandomizer>,boost::noncopyable>("ExperimentalSieveCurveRandomizer",init<>())
		.def(init<int, const vector<array<double,2>>&>())
#ifdef AGGREGPACKING_SERIALIZATION
		.def(init<int, const vector<array<double,2>>&,const string&>())
#endif
		.def("cdf",&ExperimentalSieveCurveRandomizer::cdf)
		.def("take",&ExperimentalSieveCurveRandomizer::take)
		.add_property("sizes",make_function(&ExperimentalSieveCurveRandomizer::getSizes,return_internal_reference<>()))
		.add_property("values",ESCRgetValues,&ExperimentalSieveCurveRandomizer::setValues)
		.add_property("minSize",&ExperimentalSieveCurveRandomizer::getMinSize)
		.add_property("maxSize",&ExperimentalSieveCurveRandomizer::getMaxSize)
		.add_static_property("seed",&Randomizer::getSeed,&Randomizer::setSeed)
#ifdef AGGREGPACKING_SERIALIZATION
		.add_property("fileName",&ExperimentalSieveCurveRandomizer::getFileName)
		.def("save",ESCRsave1)
		.def("save",ESCRsave2)
		.def("load",ESCRload1)
		.def("load",ESCRload2)
#endif
		;

}
/// @endcond
#endif

} // namespace
