/*
*
*      AggregPacking : Generator of random aggregate packings in concrete     
*                                                                             
*                           version 0.1 (2014-08-22)                          
* 
*                      Copyright (C)  2014  Jan Stransky                      
* 
*          Czech Technical University, Faculty of Civil Engineering,          
*      Department of Structural Mechanics, 166 29 Prague, Czech Republic      
* 
* AggregPacking is free software: you can redistribute it and/or modify it
* under the terms of the GNU Lesser General Public License as published by the
* Free Software Foundation, either version 3 of the License, or (at your
* option) any later version.
* 
* AggregPacking is distributed in the hope that it will be useful, but WITHOUT
* ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
* FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License
* for more details.
* 
* You should have received a copy of the GNU Lesser General Public License
* along with this program. If not, see <http://www.gnu.org/licenses/>.
*/

/**
 * @file
 * @brief File containing representation of different random number generators
 */

#ifndef RANDOMIZER_HPP
#define RANDOMIZER_HPP

/// @cond DOXYGEN_EXCLUDE
#define NMINVALUES_DEFAULT 10000
/// @endcond

#include <vector>
#include <array>
#include <string>
using std::vector;
using std::array;
using std::string;

#include "utils.hpp"

#ifdef AGGREGPACKING_SERIALIZATION
#include <iostream>
#include <fstream>
using std::istream;
using std::ostream;
using std::ifstream;
using std::ofstream;
#endif

#ifdef AGGREGPACKING_BOOST_PYTHON
#include <boost/python.hpp>
namespace bp = boost::python;
#endif

namespace aggregpacking {


/**********************************************************************
* RANDOMIZER
**********************************************************************/
/// Class representing most simple random numbers generator
class Randomizer {
	private:
		/// Random seed
		static int seed;

	public:
		/// Returns random number in the interval [0,1) (with uniform distribution)
		static double random() { return (double)rand()/(double)RAND_MAX; }

		/// Returns random integer in the interval [0,m-1] (with uniform istribution)
		/** @param m upper limit if the interval
		 */
		static double randint(int m) { return (int)(random()*m); }

		/// Getter for seed
		static int getSeed() { return Randomizer::seed; }

		/// Setter for seed
		/** @param s new seed
		 */
		static void setSeed(int s) { Randomizer::seed = s; srand(s); }
};



/**********************************************************************
* AGGREG SIZE RANDOMIZER
**********************************************************************/
/// Abstract class representing random size generator for given sieve curve
/** 
 */
class AggregSizeRandomizer: public Randomizer {
	public:
		/// typedef for sizes type
		typedef vector<double> sizes_t;

	protected:
		/// Array of increasing sizes. Discrete representation of given sieve curve in terms of aggregate size
		sizes_t sizes;

		/// According to given parameters, constructs sizes. It is numerically converted sieve curve from terms of cumulative volume to the terms of sizes
		void initSizes(unsigned int nMinValues);

#ifdef AGGREGPACKING_SERIALIZATION
		/// File name for saving itself
		string fileName;

		/// Save sizes to ostream
		/** @param os output stream
		 */
		void saveSizes(ostream& os) const;

		/// Load sizes from istream
		/** @param is input stream
		 */
		void loadSizes(istream& is);

		/// Create and return a new AggregSizeRandomizer from given istream
		/** @param is input stream
		*/
		static AggregSizeRandomizer* Load(istream& is);

#endif
	public:
#ifdef AGGREGPACKING_SERIALIZATION
		/// Default constructor
		AggregSizeRandomizer() : AggregSizeRandomizer(string()) {};

		/// Constructor from file name
		/** @param m file name
		 */
		AggregSizeRandomizer(const string& m) : fileName(m) {};
#else
		AggregSizeRandomizer() {};
#endif

		/** Cumulative relative mass function. Returns what is the total mass of all aggregates spmaller than given size (normalized with total mass of all aggregates)
		 * @param size given size
		 * @return relative cumulative mass of all particles smaller than given size
		 */
		virtual double cdf(double size) const = 0;

		/// Returns random size. A lot of random sizes should converge to the prescribed sieve curve. Mass/volume of aggregate is assumed pow(size,3)
		double take() const;

		/// Getter for sizes
		const sizes_t& getSizes() { return sizes; }

		/// Returns minimum defined size
		virtual double getMinSize() const { return sizes[sizes.size()-1]; }

		/// Returns maximum defined size
		virtual double getMaxSize() const { return sizes[0]; }

#ifdef AGGREGPACKING_SERIALIZATION
		/// Getter for file name
		string getFileName() { return fileName; }

		/// Returns string representation of Sphere class name
		virtual const string& getClassName() = 0;

		/// Saves receiver into specified file
		/** @param fName given file name
		 */
		virtual void save(const string& fName) const { saveObject(*this,fName,fileName); }

		/// Saves receiver into this->fileName
		virtual void save() const { save(string()); }

		/// Loads receiver from specified file
		/** @param fName given file name
		 *  @return true if successful, false otherwise
		 */
		virtual bool load(const string& fName) { return loadObject(*this,fName,fileName); }

		/// Loads receiver from this->fileName
		/** @return true if successful, false otherwise
		 */
		virtual bool load() { load(string()); }

		/// AggregSizeRandomizer to ostream shift operator
		friend ostream& operator<<(ostream& f, const AggregSizeRandomizer& o);

		/// AggregSizeRandomizer from istream shift operator
		friend istream& operator>>(istream& f, AggregSizeRandomizer& o);
#endif
};






/**********************************************************************
* FULLER CURVE RANDOMIZER
**********************************************************************/
/// Random size generator for Fuller curve
/** Fuller curve is defined as P(d) = (d/dmax)^n. Exponent n is usually between 0.4-0.7, most often 0.5. Apart from dmax (maximum aggrgate size), also minimum aggregate size has to be defined
 */
class FullerCurveRandomizer : public AggregSizeRandomizer {
	protected:
		/// Exponent of analytical expression
		double exponent;

		/// User defined minimum aggregate size
		double minSize;

		/// Maximum size from the analytical expression
		double maxSize;

	public:
		/// Default constructor
		/** sets exponent=0.5, minSize=0.5e-3, maxSize=16e-3 */
		FullerCurveRandomizer() : FullerCurveRandomizer(NMINVALUES_DEFAULT,0.5,.5e-3,16e-3) {}

#ifdef AGGREGPACKING_SERIALIZATION
		/// Constructor
		/** @param n minimum length of sizes attribute
		 *  @param e exponent
		 *  @param minSize minimum aggregate size
		 *  @param maxSize maximum aggregate size
		 */
		FullerCurveRandomizer(int n, double e, double minSize, double maxSize) : FullerCurveRandomizer(n,e,minSize,maxSize,string()) {}

		/// Constructor
		/** @param n minimum length of sizes attribute
		 *  @param e exponent
		 *  @param minSize minimum aggregate size
		 *  @param maxSize maximum aggregate size
		 *  @param m file name
		 */
		FullerCurveRandomizer(int n, double e, double minSize, double maxSize, const string& m) : AggregSizeRandomizer(m), exponent(e), minSize(minSize), maxSize(maxSize) {
			initSizes(n);
		}
#else
		FullerCurveRandomizer(int n, double e, double minSize, double maxSize) : AggregSizeRandomizer(), exponent(e), minSize(minSize), maxSize(maxSize) {
			initSizes(n);
		}
#endif

		virtual double cdf(double size) const;

		/// Getter for exponent
		double getExponent() { return exponent; }

		/// Setter for exponent
		void setExponent(double v) { exponent = v; }

		virtual double getMinSize() const { return minSize; }

		virtual double getMaxSize() const { return maxSize; }

#ifdef AGGREGPACKING_SERIALIZATION
		/// String representation of FullerCurveRandomizer class name
		const static string className;

		/// Returns string representation of FullerCurveRandomizer class name
		virtual const string& getClassName() { return FullerCurveRandomizer::className; }

		/// FullerCurveRandomizer to ostream shift operator
		friend ostream& operator<<(ostream& f, const FullerCurveRandomizer& o);

		/// FullerCurveRandomizer from istream shift operator
		friend istream& operator>>(istream& f, FullerCurveRandomizer& o);
#endif
};
//**********************************************************************



/**********************************************************************
* EXPERIMENTAL SIEVE CURVE RANDOMIZER
**********************************************************************/
/// Random size generator for experimental sieve curve
/** Define by pair array (size,relative cumulative mass)
 */
class ExperimentalSieveCurveRandomizer : public AggregSizeRandomizer {
	public:
		/// typedef for sieve curve
		typedef vector<array<double,2>> sieve_curve;

	protected:
		/// Sieve curve as a pair array (size,relative cumulative mass)
		sieve_curve values;

		/// Checks, if given sieve curve is correct. Throw error if not
		void check();

		/// Finds an interval which size blongs to
		/** @param size given size
		 * @return number of interval which size blongs to
		 */
		int findInterval(double size) const;

	public:
		/// Default constructor, sets values to ((0.5e-3,0.0),(16e-3,1.0))
		ExperimentalSieveCurveRandomizer() : ExperimentalSieveCurveRandomizer(NMINVALUES_DEFAULT,{{.5e-3,0.},{16e-3,1.}}) {};

#ifdef AGGREGPACKING_SERIALIZATION
		/// Constructor from nMinValues and sieve_curve
		/** @param n minimum number of sizes
		 * @param v sieve curve
		 */
		ExperimentalSieveCurveRandomizer(int n, const sieve_curve& v) : ExperimentalSieveCurveRandomizer(n,v,string()) {};

		/// Constructor from nMinValues, sieve_curve and file name
		/** @param n minimum number of sizes
		 * @param v sieve curve
		 * @param m file name
		 */
		ExperimentalSieveCurveRandomizer(int n, const sieve_curve& v, const string& m) : AggregSizeRandomizer(m), values(v) {
			check();
			initSizes(n);
		}
#else
		ExperimentalSieveCurveRandomizer(int n, const sieve_curve& v) : AggregSizeRandomizer(), values(v) {
			check();
			initSizes(n);
		}
#endif

		/// Getter for values (returns by const reference)
		const sieve_curve& getValues() const { return values; }

		/// Getter for values (returns by value)
		sieve_curve getValues() { return values; }

		/// Setter for values
		void setValues(const sieve_curve& sc) { values = sc; check(); }

		virtual double cdf(double size) const;

		virtual double getMinSize() const { return values[0][0]; }

		virtual double getMaxSize() const { return values[values.size()-1][0]; }

#ifdef AGGREGPACKING_SERIALIZATION
		/// String representation of ExperimentalSieveCurveRandomizer class name
		const static string className;

		/// Returns string representation of ExperimentalSieveCurveRandomizer class name
		virtual const string& getClassName() { return ExperimentalSieveCurveRandomizer::className; }

		/// ExperimentalSieveCurveRandomizer to ostream shift operator
		friend ostream& operator<<(ostream& f, const ExperimentalSieveCurveRandomizer& o);

		/// ExperimentalSieveCurveRandomizer from istream shift operator
		friend istream& operator>>(istream& f, ExperimentalSieveCurveRandomizer& o);
#endif
};
//**********************************************************************



/**********************************************************************
* PYTHON
**********************************************************************/
#ifdef AGGREGPACKING_BOOST_PYTHON
/// @cond DOXYGEN_EXCLUDE
void py_randomizer();
/// @endcond
#endif

} // namespace
#endif
